
import pytest

import wn
from wn import similarity as sim


@pytest.mark.usefixtures('mini_db')
def test_path():
    information = wn.synsets('information')[0]
    example = wn.synsets('example')[0]
    sample = wn.synsets('sample')[0]
    random_sample = wn.synsets('random sample')[0]
    datum = wn.synsets('datum')[0]
    exemplify = wn.synsets('exemplify')[0]
    assert sim.path(information, information) == 1/1
    assert sim.path(information, example) == 1/2
    assert sim.path(information, sample) == 1/3
    assert sim.path(information, random_sample) == 1/4
    assert sim.path(random_sample, datum) == 1/5
    assert sim.path(example, exemplify) == 1/4


@pytest.mark.usefixtures('mini_db')
def test_wup():
    information = wn.synsets('information')[0]
    example = wn.synsets('example')[0]
    sample = wn.synsets('sample')[0]
    random_sample = wn.synsets('random sample')[0]
    datum = wn.synsets('datum')[0]
    exemplify = wn.synsets('exemplify')[0]
    assert sim.wup(information, information) == (2*1) / (0+0+2*1)
    assert sim.wup(information, example) == (2*1) / (0+1+2*1)
    assert sim.wup(information, sample) == (2*1) / (0+2+2*1)
    assert sim.wup(information, random_sample) == (2*1) / (0+3+2*1)
    assert sim.wup(random_sample, datum) == (2*1) / (3+1+2*1)
    assert sim.wup(example, exemplify) == (2*1) / (2+1+2*1)




import pytest

import wn


@pytest.mark.usefixtures('empty_db')
def test_lexicons_empty():
    assert len(wn.lexicons()) == 0


@pytest.mark.usefixtures('mini_db')
def test_lexicons_mini():
    assert len(wn.lexicons()) == 2
    assert all(isinstance(lex, wn.Lexicon) for lex in wn.lexicons())

    results = wn.lexicons(lang='en')
    assert len(results) == 1 and results[0].language == 'en'
    results = wn.lexicons(lang='es')
    assert len(results) == 1 and results[0].language == 'es'

    results = wn.lexicons(lexicon='*')
    assert len(results) == 2
    results = wn.lexicons(lexicon='test-en')
    assert len(results) == 1 and results[0].language == 'en'
    results = wn.lexicons(lexicon='test-en:1')
    assert len(results) == 1 and results[0].language == 'en'
    results = wn.lexicons(lexicon='test-en:*')
    assert len(results) == 1 and results[0].language == 'en'


@pytest.mark.usefixtures('mini_db')
def test_lexicons_unknown():
    results = wn.lexicons(lang='unk')
    assert len(results) == 0
    results = wn.lexicons(lexicon='test-unk')
    assert len(results) == 0


@pytest.mark.usefixtures('empty_db')
def test_words_empty():
    assert len(wn.words()) == 0


@pytest.mark.usefixtures('mini_db')
def test_words_mini():
    assert len(wn.words()) == 14
    assert all(isinstance(w, wn.Word) for w in wn.words())

    words = wn.words('information')  # search lemma
    assert len(words) == 1
    assert words[0].lemma() == 'information'

    words = wn.words('exemplifies')  # search secondary form
    assert len(words) == 1
    assert words[0].lemma() == 'exemplify'

    assert len(wn.words(pos='n')) == 10
    assert all(w.pos == 'n' for w in wn.words(pos='n'))
    assert len(wn.words(pos='v')) == 4
    assert len(wn.words(pos='q')) == 0  # fake pos

    assert len(wn.words(lang='en')) == 8
    assert len(wn.words(lang='es')) == 6

    assert len(wn.words(lexicon='test-en')) == 8
    assert len(wn.words(lexicon='test-es')) == 6

    assert len(wn.words(lang='en', lexicon='test-en')) == 8
    assert len(wn.words(pos='v', lang='en')) == 2
    assert len(wn.words('information', lang='en')) == 1
    assert len(wn.words('information', lang='es')) == 0

    with pytest.raises(wn.Error):
        wn.words(lang='unk')
    with pytest.raises(wn.Error):
        wn.words(lexicon='test-unk')


@pytest.mark.usefixtures('empty_db')
def test_word_empty():
    with pytest.raises(wn.Error):
        assert wn.word('test-es-información-n')


@pytest.mark.usefixtures('mini_db')
def test_word_mini():
    assert wn.word('test-es-información-n')
    assert wn.word('test-es-información-n', lang='es')
    assert wn.word('test-es-información-n', lexicon='test-es')
    with pytest.raises(wn.Error):
        assert wn.word('test-es-información-n', lang='en')
    with pytest.raises(wn.Error):
        assert wn.word('test-es-información-n', lexicon='test-en')
    with pytest.raises(wn.Error):
        assert wn.word('test-es-información-n', lang='unk')
    with pytest.raises(wn.Error):
        assert wn.word('test-es-información-n', lexicon='test-unk')


@pytest.mark.usefixtures('empty_db')
def test_senses_empty():
    assert len(wn.senses()) == 0


@pytest.mark.usefixtures('mini_db')
def test_senses_mini():
    assert len(wn.senses()) == 14
    assert all(isinstance(s, wn.Sense) for s in wn.senses())

    senses = wn.senses('information')  # search lemma
    assert len(senses) == 1
    assert senses[0].word().lemma() == 'information'

    senses = wn.senses('exemplifies')  # search secondary form
    assert len(senses) == 1
    assert senses[0].word().lemma() == 'exemplify'

    assert len(wn.senses(pos='n')) == 10
    assert len(wn.senses(pos='v')) == 4
    assert len(wn.senses(pos='q')) == 0  # fake pos

    assert len(wn.senses(lang='en')) == 8
    assert len(wn.senses(lang='es')) == 6

    assert len(wn.senses(lexicon='test-en')) == 8
    assert len(wn.senses(lexicon='test-es')) == 6

    assert len(wn.senses(lang='en', lexicon='test-en')) == 8
    assert len(wn.senses(pos='v', lang='en')) == 2
    assert len(wn.senses('information', lang='en')) == 1
    assert len(wn.senses('information', lang='es')) == 0

    with pytest.raises(wn.Error):
        wn.senses(lang='unk')
    with pytest.raises(wn.Error):
        wn.senses(lexicon='test-unk')


@pytest.mark.usefixtures('empty_db')
def test_sense_empty():
    with pytest.raises(wn.Error):
        assert wn.sense('test-es-información-n-0001-01')


@pytest.mark.usefixtures('mini_db')
def test_sense_mini():
    assert wn.sense('test-es-información-n-0001-01')
    assert wn.sense('test-es-información-n-0001-01', lang='es')
    assert wn.sense('test-es-información-n-0001-01', lexicon='test-es')
    with pytest.raises(wn.Error):
        assert wn.sense('test-es-información-n-0001-01', lang='en')
    with pytest.raises(wn.Error):
        assert wn.sense('test-es-información-n-0001-01', lexicon='test-en')
    with pytest.raises(wn.Error):
        assert wn.sense('test-es-información-n-0001-01', lang='unk')
    with pytest.raises(wn.Error):
        assert wn.sense('test-es-información-n-0001-01', lexicon='test-unk')


@pytest.mark.usefixtures('empty_db')
def test_synsets_empty():
    assert len(wn.synsets()) == 0


@pytest.mark.usefixtures('mini_db')
def test_synsets_mini():
    assert len(wn.synsets()) == 10
    assert all(isinstance(ss, wn.Synset) for ss in wn.synsets())

    synsets = wn.synsets('information')  # search lemma
    assert len(synsets) == 1
    assert 'information' in synsets[0].lemmas()

    synsets = wn.synsets('exemplifies')  # search secondary form
    assert len(synsets) == 1
    assert 'exemplify' in synsets[0].lemmas()

    assert len(wn.synsets(pos='n')) == 8
    assert len(wn.synsets(pos='v')) == 2
    assert len(wn.synsets(pos='q')) == 0  # fake pos

    assert len(wn.synsets(ili='i67469')) == 2
    assert len(wn.synsets(ili='i67468')) == 0

    assert len(wn.synsets(lang='en')) == 6
    assert len(wn.synsets(lang='es')) == 4

    assert len(wn.synsets(lexicon='test-en')) == 6
    assert len(wn.synsets(lexicon='test-es')) == 4

    assert len(wn.synsets(lang='en', lexicon='test-en')) == 6
    assert len(wn.synsets(pos='v', lang='en')) == 1
    assert len(wn.synsets('information', lang='en')) == 1
    assert len(wn.synsets('information', lang='es')) == 0
    assert len(wn.synsets(ili='i67469', lang='es')) == 1

    with pytest.raises(wn.Error):
        wn.synsets(lang='unk')
    with pytest.raises(wn.Error):
        wn.synsets(lexicon='test-unk')


@pytest.mark.usefixtures('empty_db')
def test_synset_empty():
    with pytest.raises(wn.Error):
        assert wn.synset('test-es-0001-n')


@pytest.mark.usefixtures('mini_db')
def test_synset_mini():
    assert wn.synset('test-es-0001-n')
    assert wn.synset('test-es-0001-n', lang='es')
    assert wn.synset('test-es-0001-n', lexicon='test-es')
    with pytest.raises(wn.Error):
        assert wn.synset('test-es-0001-n', lang='en')
    with pytest.raises(wn.Error):
        assert wn.synset('test-es-0001-n', lexicon='test-en')
    with pytest.raises(wn.Error):
        assert wn.synset('test-es-0001-n', lang='unk')
    with pytest.raises(wn.Error):
        assert wn.synset('test-es-0001-n', lexicon='test-unk')




import pytest

import wn


@pytest.mark.usefixtures('mini_db')
def test_word_senses():
    assert len(wn.word('test-en-information-n').senses()) == 1
    assert len(wn.word('test-es-información-n').senses()) == 1


@pytest.mark.usefixtures('mini_db')
def test_word_synsets():
    assert len(wn.word('test-en-information-n').synsets()) == 1
    assert len(wn.word('test-es-información-n').synsets()) == 1


@pytest.mark.usefixtures('mini_db')
def test_word_translate():
    assert len(wn.word('test-en-example-n').translate(lang='es')) == 1
    assert len(wn.word('test-es-ejemplo-n').translate(lang='en')) == 1


@pytest.mark.usefixtures('mini_db')
def test_sense_word():
    assert (wn.sense('test-en-information-n-0001-01').word()
            == wn.word('test-en-information-n'))
    assert (wn.sense('test-es-información-n-0001-01').word()
            == wn.word('test-es-información-n'))


@pytest.mark.usefixtures('mini_db')
def test_sense_synset():
    assert (wn.sense('test-en-information-n-0001-01').synset()
            == wn.synset('test-en-0001-n'))
    assert (wn.sense('test-es-información-n-0001-01').synset()
            == wn.synset('test-es-0001-n'))


@pytest.mark.usefixtures('mini_db')
def test_sense_examples():
    assert wn.sense('test-en-information-n-0001-01').examples() == []
    assert wn.sense('test-es-información-n-0001-01').examples() == []


@pytest.mark.usefixtures('mini_db')
def test_sense_lexicalized():
    assert wn.sense('test-en-information-n-0001-01').lexicalized()
    assert wn.sense('test-es-información-n-0001-01').lexicalized()


@pytest.mark.usefixtures('mini_db')
def test_sense_translate():
    assert len(wn.sense('test-en-information-n-0001-01').translate(lang='es')) == 1
    assert len(wn.sense('test-es-información-n-0001-01').translate(lang='en')) == 1


@pytest.mark.usefixtures('mini_db')
def test_synset_senses():
    assert len(wn.synset('test-en-0003-v').senses()) == 2
    assert len(wn.synset('test-es-0003-v').senses()) == 2


@pytest.mark.usefixtures('mini_db')
def test_synset_words():
    assert len(wn.synset('test-en-0003-v').words()) == 2
    assert len(wn.synset('test-es-0003-v').words()) == 2


@pytest.mark.usefixtures('mini_db')
def test_synset_lemmas():
    assert wn.synset('test-en-0003-v').lemmas() == ['exemplify', 'illustrate']
    assert wn.synset('test-es-0003-v').lemmas() == ['ejemplificar', 'ilustrar']


@pytest.mark.usefixtures('mini_db')
def test_synset_definition():
    assert wn.synset('test-en-0001-n').definition() == 'something that informs'
    assert wn.synset('test-es-0001-n').definition() == 'algo que informa'


@pytest.mark.usefixtures('mini_db')
def test_synset_examples():
    assert wn.synset('test-en-0001-n').examples() == ['"this is information"']
    assert wn.synset('test-es-0001-n').examples() == ['"este es la información"']


@pytest.mark.usefixtures('mini_db')
def test_synset_lexicalized():
    assert wn.synset('test-en-0001-n').lexicalized()
    assert wn.synset('test-es-0001-n').lexicalized()


@pytest.mark.usefixtures('mini_db')
def test_synset_translate():
    assert len(wn.synset('test-en-0001-n').translate(lang='es')) == 1
    assert len(wn.synset('test-es-0001-n').translate(lang='en')) == 1




from wn import lmf


class TestLexicon:

    def test_entry_ids(self):
        lex = lmf.Lexicon(
            'id', 'Test', 'tst', 'test@example.com', 'CC-BY 4.0', '1',
            lexical_entries=[],
            synsets=[])
        assert lex.entry_ids() == set()


def test_is_lmf(datadir):
    assert lmf.is_lmf(datadir / 'mini-lmf-1.0.xml')
    assert lmf.is_lmf(str(datadir / 'mini-lmf-1.0.xml'))
    assert not lmf.is_lmf(datadir / 'README.md')
    assert not lmf.is_lmf(datadir / 'missing.xml')


def test_load(mini_lmf_1_0):
    lexicons = lmf.load(mini_lmf_1_0)
    assert len(lexicons) == 2
    lexicon = lexicons[0]

    assert lexicon.id == 'test-en'
    assert lexicon.label == 'Testing English WordNet'
    assert lexicon.language == 'en'
    assert lexicon.email == 'maintainer@example.com'
    assert lexicon.license == 'https://creativecommons.org/licenses/by/4.0/'
    assert lexicon.version == '1'
    assert lexicon.url == 'https://example.com/test-en'

    assert len(lexicon.lexical_entries) == 8
    le = lexicon.lexical_entries[0]
    assert le.id == 'test-en-information-n'

    assert le.lemma.form == 'information'
    assert le.lemma.pos == 'n'
    assert le.lemma.script is None
    assert len(le.lemma.tags) == 0

    assert len(le.forms) == 0

    assert len(le.senses) == 1
    assert le.senses[0].id == 'test-en-information-n-0001-01'
    assert le.senses[0].synset == 'test-en-0001-n'
    assert len(le.senses[0].relations) == 0
    # assert le.senses[0].relations[0].target == 'test-en-exemplify-v-01023137-01'
    # assert le.senses[0].relations[0].type == 'derivation'

    assert len(lexicon.synsets) == 6

    assert lexicons[1].id == 'test-es'


def test_dump(mini_lmf_1_0, tmp_path):
    lexicons = lmf.load(mini_lmf_1_0)
    tmpdir = tmp_path / 'test_dump'
    tmpdir.mkdir()
    lmf.dump(lexicons, tmpdir / 'mini_lmf_dump.xml')




import math

import wn
from wn._core import Synset


def path(synset1, synset2):
    """Return the Path similarity of *synset1* and *synset2*.

    When :math:`d` is the length of the shortest path from *synset1*
    to *synset2*, the path similarity is: :math:`\\frac{1}{d + 1}`

    """
    distance = len(synset1.shortest_path(synset2, simulate_root=True))
    return 1 / (distance + 1)


def wup(synset1: Synset, synset2: Synset) -> float:
    """Return the Wu-Palmer similarity of *synset1* and *synset2*.

    When *lch* is the lowest common hypernym for *synset1* and
    *synset2*, *n1* is the shortest path distance from *synset1* to
    *lch*, *n2* is the shortest path distance from *synset2* to *lch*,
    and *n3* is the number of nodes (distance + 1) from *lch* to the
    root node, then the Wu-Palmer similarity is:
    :math:`\\frac{2(n3)}{n1 + n2 + 2(n3)}`

    """
    lch = synset1.lowest_common_hypernyms(synset2, simulate_root=True)[0]
    n3 = lch.max_depth() + 1
    n1 = len(synset1.shortest_path(lch, simulate_root=True))
    n2 = len(synset2.shortest_path(lch, simulate_root=True))
    return (2 * n3) / (n1 + n2 + 2 * n3)


def lch(synset1: Synset, synset2: Synset, max_depth: int = 0) -> float:
    """Return the Leacock-Chodorow similarity of *synset1* and *synset2*."""
    distance = len(synset1.shortest_path(synset2, simulate_root=True))
    if max_depth <= 0:
        raise wn.Error('max_depth must be greater than 0')
    return -math.log((distance + 1) / (2 * max_depth))




from typing import Optional, Type
from pathlib import Path
import logging

import requests

import wn
from wn._util import is_url
from wn.util import ProgressHandler, ProgressBar
from wn._add import add as add_to_db
from wn import config


CHUNK_SIZE = 8 * 1024  # how many KB to read at a time
TIMEOUT = 10  # number of seconds to wait for a server response


logger = logging.getLogger('wn')


def download(
        project_or_url: str,
        add: bool = True,
        progress_handler: Optional[Type[ProgressHandler]] = ProgressBar,
) -> Path:
    """Download the resource specified by *project_or_url*.

    First the URL of the resource is determined and then, depending on
    the parameters, the resource is downloaded and added to the
    database.  The function then returns the path of the cached file.

    If *project_or_url* starts with `'http://'` or `'https://'`, then
    it is taken to be the URL for the resource. Otherwise,
    *project_or_url* is taken as a :ref:`project specifier
    <lexicon-specifiers>` and the URL is taken from a matching entry
    in Wn's project index. If no project matches the specifier,
    :exc:`wn.Error` is raised.

    If the URL has been downloaded and cached before, the cached file
    is used. Otherwise the URL is retrieved and stored in the cache.

    If the *add* paramter is ``True`` (default), the downloaded
    resource is added to the database.

    >>> wn.download('ewn:2020')
    Added ewn:2020 (English WordNet)

    The *progress_handler* parameter takes a callable that is called
    after every chunk of bytes is received. The handler function
    should have the following signature:

    .. code-block:: python

       def progress_handler(n: int, **kwargs) -> str:
           ...

    The *n* parameter is the number of bytes received in a chunk. When
    a request is sent, the handler function is called with ``max`` key
    in *kwargs* mapped to the total number of bytes (taken from the
    response's ``Content-Length`` header). A ``status`` key may also
    indicate the current status (``Requesting``, ``Receiving``,
    ``Completed``).

    """
    if is_url(project_or_url):
        url = project_or_url
    else:
        info = config.get_project_info(project_or_url)
        url = info['resource_url']
    logger.info('download url: %s', url)

    path = config.get_cache_path(url)
    logger.info('download cache path: %s', path)

    if progress_handler is None:
        progress_handler = ProgressHandler
    progress = progress_handler(message='Download', unit='bytes')

    try:
        if path.exists():
            progress.flash(f'Cached file found: {path!s}')
        else:
            _download(url, path, progress)
    finally:
        progress.close()

    if add:
        try:
            add_to_db(path, progress_handler=progress_handler)
        except wn.Error as exc:
            raise wn.Error(
                f'could not add downloaded file: {path}\n  You might try '
                'deleting the cached file and trying the download again.'
            ) from exc

    return path


def _download(url: str, path: Path, progress: ProgressHandler) -> None:
    size: int = 0
    try:
        with open(path, 'wb') as f:
            progress.set(status='Requesting')
            with requests.get(url, stream=True, timeout=TIMEOUT) as response:
                size = int(response.headers.get('Content-Length', 0))
                progress.set(total=size, status='Receiving')
                for chunk in response.iter_content(chunk_size=CHUNK_SIZE):
                    if chunk:
                        f.write(chunk)
                    progress.update(len(chunk))
                progress.set(status='Complete')
    except requests.exceptions.RequestException as exc:
        path.unlink()
        raise wn.Error(f'Download failed at {size} bytes') from exc
    except KeyboardInterrupt as exc:
        path.unlink()
        raise wn.Error(f'Download cancelled at {size} bytes') from exc
    except Exception:
        path.unlink()
        raise



"""
Adding and removing lexicons to/from the database.
"""

from typing import Optional, Type
import logging

import wn
from wn._types import AnyPath
from wn._db import connect
from wn._queries import find_lexicons
from wn.util import ProgressHandler, ProgressBar
from wn.project import iterpackages
from wn import lmf


logger = logging.getLogger('wn')


BATCH_SIZE = 1000

POS_QUERY = '''
    SELECT p.rowid
      FROM parts_of_speech AS p
     WHERE p.pos = ?
'''
ENTRY_QUERY = '''
    SELECT e.rowid
      FROM entries AS e
     WHERE e.id = ?
       AND e.lexicon_rowid = ?
'''
SENSE_QUERY = '''
    SELECT s.rowid
      FROM senses AS s
     WHERE s.id = ?
       AND s.lexicon_rowid = ?
'''
SYNSET_QUERY = '''
    SELECT ss.rowid
      FROM synsets AS ss
     WHERE ss.id = ?
       AND ss.lexicon_rowid = ?
'''


def add(
    source: AnyPath,
    progress_handler: Optional[Type[ProgressHandler]] = ProgressBar,
) -> None:
    """Add the LMF file at *source* to the database.

    The file at *source* may be gzip-compressed or plain text XML.

    >>> wn.add('english-wordnet-2020.xml')
    Added ewn:2020 (English WordNet)

    The *progress_handler* parameter takes a callable that is called
    after every block of rows is inserted. The handler function
    should have the following signature:

    .. code-block:: python

       def progress_handler(n: int, **kwargs) -> str:
           ...

    The *n* parameter is the number of rows last inserted into the
    database.  A ``status`` key on the *kwargs* indicates the current
    status of adding the lexicon (``Inspecting``, ``ILI``, ``Synset``,
    etc.). After inspecting the file, a ``max`` keyword on *kwargs*
    indicates the total number of rows to insert.
    """
    if progress_handler is None:
        progress_handler = ProgressHandler
    progress = progress_handler(message='Database', unit='\b')

    logger.info('adding project to database')
    logger.info('  database: %s', wn.config.database_path)
    logger.info('  project file: %s', source)

    try:
        for package in iterpackages(source):
            _add_lmf(package.resource_file(), progress)
    finally:
        progress.close()


def _add_lmf(
    source,
    progress: ProgressHandler,
) -> None:
    with connect() as conn:
        cur = conn.cursor()
        # these two settings increase the risk of database corruption
        # if the system crashes during a write, but they should also
        # make inserts much faster
        cur.execute('PRAGMA synchronous = OFF')
        cur.execute('PRAGMA journal_mode = MEMORY')

        # abort if any lexicon in *source* is already added
        progress.flash(f'Checking {source!s}')
        all_infos = list(_precheck(source, cur))

        if not all_infos:
            progress.flash(f'{source}: No lexicons found')
            return
        elif all(info.get('skip', False) for info in all_infos):
            progress.flash(f'{source}: Some or all lexicons already added')
            return

        # all clear, try to add them
        progress.flash(f'Reading {source!s}')
        for lexicon, info in zip(lmf.load(source), all_infos):

            if info.get('skip', False):
                progress.flash(
                    f'Skipping {info["id"]:info["version"]} ({info["label"]})\n',
                )
                continue

            sense_ids = lexicon.sense_ids()
            synset_ids = lexicon.synset_ids()

            cur.execute(
                'INSERT INTO lexicons VALUES (null,?,?,?,?,?,?,?,?,?)',
                (lexicon.id,
                 lexicon.label,
                 lexicon.language,
                 lexicon.email,
                 lexicon.license,
                 lexicon.version,
                 lexicon.url,
                 lexicon.citation,
                 lexicon.meta))
            lexid = cur.lastrowid

            counts = info['counts']
            count = sum(counts.get(name, 0) for name in
                        ('LexicalEntry', 'Lemma', 'Form',  # 'Tag',
                         'Sense', 'SenseRelation', 'Example',  # 'Count',
                         # 'SyntacticBehaviour',
                         'Synset', 'Definition',  # 'ILIDefinition',
                         'SynsetRelation'))
            count += counts.get('Synset', 0)  # again for ILIs
            progress.set(count=0, total=count)

            synsets = lexicon.synsets
            entries = lexicon.lexical_entries

            _insert_ilis(synsets, cur, progress)
            _insert_synsets(synsets, lexid, cur, progress)
            _insert_entries(entries, lexid, cur, progress)
            _insert_forms(entries, lexid, cur, progress)
            _insert_senses(entries, lexid, cur, progress)

            _insert_synset_relations(synsets, lexid, cur, progress)
            _insert_sense_relations(entries, lexid, 'sense_relations',
                                    sense_ids, cur, progress)
            _insert_sense_relations(entries, lexid, 'sense_synset_relations',
                                    synset_ids, cur, progress)

            _insert_synset_definitions(synsets, lexid, cur, progress)
            _insert_examples([sense for entry in entries for sense in entry.senses],
                             lexid, 'sense_examples', cur, progress)
            _insert_examples(synsets, lexid, 'synset_examples', cur, progress)
            progress.set(status='')  # clear type string
            progress.flash(f'Added {lexicon.id}:{lexicon.version} ({lexicon.label})')


def _precheck(source, cur):
    for info in lmf.scan_lexicons(source):
        id = info['id']
        version = info['version']
        if cur.execute(
            'SELECT * FROM lexicons WHERE id = ? AND version = ?',
            (id, version)
        ).fetchone():
            info['skip'] = True
        yield info


def _split(sequence):
    i = 0
    for j in range(0, len(sequence), BATCH_SIZE):
        yield sequence[i:j]
        i = j
    yield sequence[i:]


def _insert_ilis(synsets, cur, progress):
    progress.set(status='ILI')
    for batch in _split(synsets):
        data = (
            (synset.ili,
             synset.ili_definition.text if synset.ili_definition else None,
             synset.ili_definition.meta if synset.ili_definition else None)
            for synset in batch if synset.ili and synset.ili != 'in'
        )
        cur.executemany('INSERT OR IGNORE INTO ilis VALUES (?,?,?)', data)
        progress.update(len(batch))


def _insert_synsets(synsets, lex_id, cur, progress):
    progress.set(status='Synsets')
    query = f'INSERT INTO synsets VALUES (null,?,?,?,({POS_QUERY}),?,?)'
    for batch in _split(synsets):
        data = (
            (synset.id,
             lex_id,
             synset.ili if synset.ili and synset.ili != 'in' else None,
             synset.pos,
             # lexfile_map.get(synset.meta.subject) if synset.meta else None,
             synset.lexicalized,
             synset.meta)
            for synset in batch
        )
        cur.executemany(query, data)
        progress.update(len(batch))


def _insert_synset_definitions(synsets, lexid, cur, progress):
    progress.set(status='Definitions')
    query = f'INSERT INTO definitions VALUES (({SYNSET_QUERY}),?,?,?)'
    for batch in _split(synsets):
        data = [
            (synset.id, lexid,
             definition.text,
             definition.language,
             # definition.source_sense,
             # lexid,
             definition.meta)
            for synset in batch
            for definition in synset.definitions
        ]
        cur.executemany(query, data)
        progress.update(len(data))


def _insert_synset_relations(synsets, lexid, cur, progress):
    progress.set(status='Synset Relations')
    type_query = 'SELECT r.rowid FROM synset_relation_types AS r WHERE r.type = ?'
    query = f'''
        INSERT INTO synset_relations
        VALUES (({SYNSET_QUERY}),
                ({SYNSET_QUERY}),
                ({type_query}),
                ?)
    '''
    for batch in _split(synsets):
        data = [
            (synset.id, lexid,
             relation.target, lexid,
             relation.type,
             relation.meta)
            for synset in batch
            for relation in synset.relations
        ]
        cur.executemany(query, data)
        progress.update(len(data))


def _insert_entries(entries, lex_id, cur, progress):
    progress.set(status='Words')
    query = f'INSERT INTO entries VALUES (null,?,?,({POS_QUERY}),?)'
    for batch in _split(entries):
        data = (
            (entry.id,
             lex_id,
             entry.lemma.pos,
             entry.meta)
            for entry in batch
        )
        cur.executemany(query, data)
        progress.update(len(batch))


def _insert_forms(entries, lexid, cur, progress):
    progress.set(status='Word Forms')
    query = f'INSERT INTO forms VALUES (null,({ENTRY_QUERY}),?,?,?)'
    for batch in _split(entries):
        forms = []
        for entry in batch:
            forms.append((entry.id, lexid, entry.lemma.form, entry.lemma.script, 0))
            forms.extend((entry.id, lexid, form.form, form.script, i)
                         for i, form in enumerate(entry.forms, 1))
        cur.executemany(query, forms)
        progress.update(len(forms))


def _insert_senses(entries, lexid, cur, progress):
    progress.set(status='Senses')
    query = f'''
        INSERT INTO senses
        VALUES (null,
                ?,
                ?,
                ({ENTRY_QUERY}),
                ?,
                ({SYNSET_QUERY}),
                ?,
                ?)
    '''
    for batch in _split(entries):
        data = [
            (sense.id,
             lexid,
             entry.id, lexid,
             i,
             sense.synset, lexid,
             # sense.meta.identifier if sense.meta else None,
             # adjmap.get(sense.adjposition),
             sense.lexicalized,
             sense.meta)
            for entry in batch
            for i, sense in enumerate(entry.senses)
        ]
        cur.executemany(query, data)
        progress.update(len(data))


def _insert_sense_relations(entries, lexid, table, ids, cur, progress):
    progress.set(status='Sense Relations')
    target_query = SENSE_QUERY if table == 'sense_relations' else SYNSET_QUERY
    type_query = 'SELECT r.rowid FROM sense_relation_types AS r WHERE r.type = ?'
    query = f'''
        INSERT INTO {table}
        VALUES (({SENSE_QUERY}),
                ({target_query}),
                ({type_query}),
                ?)
    '''
    for batch in _split(entries):
        data = [
            (sense.id, lexid,
             relation.target, lexid,
             relation.type,
             relation.meta)
            for entry in batch
            for sense in entry.senses
            for relation in sense.relations if relation.target in ids
        ]
        # be careful of SQL injection here
        cur.executemany(query, data)
        progress.update(len(data))


def _insert_examples(objs, lexid, table, cur, progress):
    progress.set(status='Examples')
    query = f'INSERT INTO {table} VALUES (({SYNSET_QUERY}),?,?,?)'
    for batch in _split(objs):
        data = [
            (obj.id, lexid,
             example.text,
             example.language,
             example.meta)
            for obj in batch
            for example in obj.examples
        ]
        # be careful of SQL injection here
        cur.executemany(query, data)
        progress.update(len(data))


def remove(lexicon: str) -> None:
    """Remove lexicon(s) from the database.

    The *lexicon* argument is a :ref:`lexicon specifier
    <lexicon-specifiers>`. Note that this removes a lexicon and not a
    project, so the lexicons of projects containing multiple lexicons
    will need to be removed individually.

    >>> wn.remove('ewn:2019')

    """
    with connect() as conn:
        for rowid, id, _, _, _, _, version, *_ in find_lexicons(lexicon=lexicon):
            conn.execute('DELETE FROM entries WHERE lexicon_rowid = ?', (rowid,))
            conn.execute('DELETE FROM synsets WHERE lexicon_rowid = ?', (rowid,))
            conn.execute('DELETE FROM syntactic_behaviours WHERE lexicon_rowid = ?',
                         (rowid,))
            conn.execute('DELETE FROM lexicons WHERE rowid = ?', (rowid,))




from typing import TypeVar, Optional, List, Tuple, Dict, Set, Iterator

import wn
from wn._types import Metadata
from wn._util import flatten
from wn._db import NON_ROWID
from wn._queries import (
    find_lexicons,
    find_entries,
    find_senses,
    find_synsets,
    get_lexicon,
    get_entry_senses,
    get_sense_relations,
    get_sense_synset_relations,
    get_synset_relations,
    get_synset_members,
    get_synsets_for_ilis,
    get_examples,
    get_definitions,
    get_metadata,
    get_lexicalized,
)

_FAKE_ROOT = '*ROOT*'
_INFERRED_SYNSET = '*INFERRED*'


class _DatabaseEntity:
    __slots__ = '_id',

    _ENTITY_TYPE = ''

    def __init__(self, _id: int = NON_ROWID):
        self._id = _id        # Database-internal id (e.g., rowid)

    def __eq__(self, other):
        if not isinstance(other, _DatabaseEntity):
            return NotImplemented
        # the _id of different kinds of entities, such as Synset and
        # Sense, can be the same, so make sure they are the same type
        # of object first
        return (self._ENTITY_TYPE == other._ENTITY_TYPE
                and self._id == other._id)

    def __lt__(self, other):
        if not isinstance(other, _DatabaseEntity):
            return NotImplemented
        elif self._ENTITY_TYPE != other._ENTITY_TYPE:
            return NotImplemented
        else:
            return self._id < other._id

    def __hash__(self):
        return hash((self._ENTITY_TYPE, self._id))


class Lexicon(_DatabaseEntity):
    """A class representing a wordnet lexicon.

    Attributes:
        id: The lexicon's identifier.
        label: The full name of lexicon.
        language: The BCP 47 language code of lexicon.
        email: The email address of the wordnet maintainer.
        license: The URL or name of the wordnet's license.
        version: The version string of the resource.
        url: The project URL of the wordnet.
        citation: The canonical citation for the project.
    """
    __slots__ = ('id', 'label', 'language', 'email', 'license',
                 'version', 'url', 'citation')
    __module__ = 'wn'

    _ENTITY_TYPE = 'lexicons'

    def __init__(
            self,
            id: str,
            label: str,
            language: str,
            email: str,
            license: str,
            version: str,
            url: str = None,
            citation: str = None,
            _id: int = NON_ROWID,
    ):
        super().__init__(_id=_id)
        self.id = id
        self.label = label
        self.language = language
        self.email = email
        self.license = license
        self.version = version
        self.url = url
        self.citation = citation

    def __repr__(self):
        id, ver, lg = self.id, self.version, self.language
        return f'<Lexicon {id}:{ver} [{lg}]>'

    def metadata(self) -> Metadata:
        """Return the lexicon's metadata."""
        return get_metadata(self._id, 'lexicons')

    def specifier(self) -> str:
        """Return the *id:version* lexicon specifier."""
        return f'{self.id}:{self.version}'


class _LexiconElement(_DatabaseEntity):
    __slots__ = '_lexid', '_wordnet'

    def __init__(
            self,
            _lexid: int = NON_ROWID,
            _id: int = NON_ROWID,
            _wordnet: 'Wordnet' = None
    ):
        super().__init__(_id=_id)
        self._lexid = _lexid  # Database-internal lexicon id
        self._wordnet = _wordnet

    def lexicon(self):
        return _to_lexicon(get_lexicon(self._lexid))


class Form(str):
    """A word-form string with additional attributes."""
    __slots__ = '_id', 'script',
    __module__ = 'wn'

    _id: int
    script: Optional[str]

    def __new__(cls, form: str, script: str = None, _id: int = NON_ROWID):
        obj = str.__new__(cls, form)  # type: ignore
        obj.script = script
        obj._id = _id
        return obj

    def __eq__(self, other):
        return (str.__eq__(self, other)
                and self.script == getattr(other, 'script', None))

    def __hash__(self):
        script = self.script
        if script is None:
            return str.__hash__(self)
        return hash((str(self), self.script))


class Word(_LexiconElement):
    """A class for words (also called lexical entries) in a wordnet."""
    __slots__ = 'id', 'pos', '_forms'
    __module__ = 'wn'

    _ENTITY_TYPE = 'entries'

    def __init__(
            self,
            id: str,
            pos: str,
            forms: List[Tuple[str, Optional[str], int]],
            _lexid: int = NON_ROWID,
            _id: int = NON_ROWID,
            _wordnet: 'Wordnet' = None
    ):
        super().__init__(_lexid=_lexid, _id=_id, _wordnet=_wordnet)
        self.id = id
        self.pos = pos
        self._forms = forms

    def __repr__(self) -> str:
        return f'Word({self.id!r})'

    def lemma(self) -> Form:
        """Return the canonical form of the word.

        Example:

            >>> wn.words('wolves')[0].lemma()
            'wolf'

        """
        return Form(*self._forms[0])

    def forms(self) -> List[Form]:
        """Return the list of all encoded forms of the word.

        Example:

            >>> wn.words('wolf')[0].forms()
            ['wolf', 'wolves']

        """
        return [Form(*form_data) for form_data in self._forms]

    def senses(self) -> List['Sense']:
        """Return the list of senses of the word.

        Example:

            >>> wn.words('zygoma')[0].senses()
            [Sense('ewn-zygoma-n-05292350-01')]

        """
        iterable = get_entry_senses(self._id)
        return [Sense(*sense_data, self._wordnet) for sense_data in iterable]

    def metadata(self) -> Metadata:
        """Return the word's metadata."""
        return get_metadata(self._id, 'entries')

    def synsets(self) -> List['Synset']:
        """Return the list of synsets of the word.

        Example:

            >>> wn.words('addendum')[0].synsets()
            [Synset('ewn-06411274-n')]

        """
        return [sense.synset() for sense in self.senses()]

    def derived_words(self) -> List['Word']:
        """Return the list of words linked through derivations on the senses.

        Example:

            >>> wn.words('magical')[0].derived_words()
            [Word('ewn-magic-n'), Word('ewn-magic-n')]

        """
        return [derived_sense.word()
                for sense in self.senses()
                for derived_sense in sense.get_related('derivation')]

    def translate(
        self, lexicon: str = None, *, lang: str = None,
    ) -> Dict['Sense', List['Word']]:
        """Return a mapping of word senses to lists of translated words.

        Arguments:
            lexicon: if specified, translate to words in the target lexicon(s)
            lang: if specified, translate to words with the language code

        Example:

            >>> w = wn.words('water bottle', pos='n')[0]
            >>> for sense, words in w.translate(lang='ja').items():
            ...     print(sense, [jw.lemma() for jw in words])
            ...
            Sense('ewn-water_bottle-n-04564934-01') ['水筒']

        """
        result = {}
        for sense in self.senses():
            result[sense] = [
                t_sense.word()
                for t_sense in sense.translate(lang=lang, lexicon=lexicon)
            ]
        return result


T = TypeVar('T', bound='_Relatable')


class _Relatable(_LexiconElement):
    __slots__ = 'id',

    def __init__(
            self,
            id: str,
            _lexid: int = NON_ROWID,
            _id: int = NON_ROWID,
            _wordnet: 'Wordnet' = None
    ):
        super().__init__(_lexid=_lexid, _id=_id, _wordnet=_wordnet)
        self.id = id

    def get_related(self: T, *args: str) -> List[T]:
        raise NotImplementedError

    def closure(self: T, *args: str) -> Iterator[T]:
        visited = set()
        queue = self.get_related(*args)
        while queue:
            relatable = queue.pop(0)
            if relatable.id not in visited:
                visited.add(relatable.id)
                yield relatable
                queue.extend(relatable.get_related(*args))

    def relation_paths(self: T, *args: str, end: T = None) -> Iterator[List[T]]:
        agenda: List[Tuple[List[T], Set[T]]] = [
            ([target], {self, target})
            for target in self.get_related(*args)
            if target._id != self._id  # avoid self loops?
        ]
        while agenda:
            path, visited = agenda.pop()
            if end is not None and path[-1] == end:
                yield path
            else:
                related = [target for target in path[-1].get_related(*args)
                           if target not in visited]
                if related:
                    for synset in reversed(related):
                        new_path = list(path) + [synset]
                        new_visited = visited | {synset}
                        agenda.append((new_path, new_visited))
                elif end is None:
                    yield path


class Synset(_Relatable):
    """Class for modeling wordnet synsets."""
    __slots__ = 'pos', 'ili'
    __module__ = 'wn'

    _ENTITY_TYPE = 'synsets'

    def __init__(
            self,
            id: str,
            pos: str,
            ili: str = None,
            _lexid: int = NON_ROWID,
            _id: int = NON_ROWID,
            _wordnet: 'Wordnet' = None
    ):
        super().__init__(id=id, _lexid=_lexid, _id=_id, _wordnet=_wordnet)
        self.pos = pos
        self.ili = ili

    @classmethod
    def empty(
            cls,
            id: str,
            ili: str = None,
            _lexid: int = NON_ROWID,
            _wordnet: 'Wordnet' = None
    ):
        return cls(id, pos='', ili=ili, _lexid=_lexid, _wordnet=_wordnet)

    def __hash__(self):
        # include ili and lexid in the hash so inferred synsets don't
        # hash the same
        return hash((self._ENTITY_TYPE, self.ili, self._lexid, self._id))

    def __repr__(self) -> str:
        return f'Synset({self.id!r})'

    def definition(self) -> Optional[str]:
        """Return the first definition found for the synset.

        Example:

            >>> wn.synsets('cartwheel', pos='n')[0].definition()
            'a wheel that has wooden spokes and a metal rim'

        """
        return next(
            (text for text, _, _ in get_definitions(self._id)),
            None
        )

    def examples(self) -> List[str]:
        """Return the list of examples for the synset.

        Example:

            >>> wn.synsets('orbital', pos='a')[0].examples()
            ['"orbital revolution"', '"orbital velocity"']

        """
        exs = get_examples(self._id, 'synsets')
        return [ex for ex, _, _ in exs]

    def senses(self) -> List['Sense']:
        """Return the list of sense members of the synset.

        Example:

            >>> wn.synsets('umbrella', pos='n')[0].senses()
            [Sense('ewn-umbrella-n-04514450-01')]

        """
        iterable = get_synset_members(self._id)
        return [Sense(*sense_data, self._wordnet) for sense_data in iterable]

    def lexicalized(self) -> bool:
        return get_lexicalized(self._id, 'synsets')

    def metadata(self) -> Metadata:
        """Return the synset's metadata."""
        return get_metadata(self._id, 'synsets')

    def words(self) -> List[Word]:
        """Return the list of words linked by the synset's senses.

        Example:

            >>> wn.synsets('exclusive', pos='n')[0].words()
            [Word('ewn-scoop-n'), Word('ewn-exclusive-n')]

        """
        return [sense.word() for sense in self.senses()]

    def lemmas(self) -> List[Form]:
        """Return the list of lemmas of words for the synset.

        Example:

            >>> wn.synsets('exclusive', pos='n')[0].words()
            ['scoop', 'exclusive']

        """
        return [w.lemma() for w in self.words()]

    def get_related(self, *args: str) -> List['Synset']:
        targets: List['Synset'] = []
        # first get relations from the current lexicon(s)
        if self._id != NON_ROWID:
            targets.extend(Synset(*row[2:], self._wordnet)
                           for row in get_synset_relations({self._id}, args))

        # then attempt to expand via ILI
        if self.ili is not None and self._wordnet and self._wordnet._expanded_ids:
            lexids: Tuple[int, ...]
            if self._wordnet._default_mode:
                lexids = (self._lexid,)
            else:
                lexids = self._wordnet._lexicon_ids
            expids = self._wordnet._expanded_ids

            # get expanded relation
            expss = find_synsets(ili=self.ili, lexicon_rowids=expids)
            rowids = {rowid for _, _, _, _, rowid in expss} - {self._id, NON_ROWID}
            ilis = {row[4] for row in get_synset_relations(rowids, args)} - {None}

            # map back to target lexicons
            seen = {ss._id for ss in targets}
            for row in get_synsets_for_ilis(ilis, lexicon_rowids=lexids):
                if row[-1] not in seen:
                    targets.append(Synset(*row, self._wordnet))

            # add empty synsets for ILIs without a target in lexids
            for ili in (ilis - {tgt.ili for tgt in targets}):
                targets.append(
                    Synset.empty(
                        id=_INFERRED_SYNSET,
                        ili=ili,
                        _lexid=self._lexid,
                        _wordnet=self._wordnet
                    )
                )

        return targets

    def _hypernym_paths(
            self, simulate_root: bool, include_self: bool
    ) -> List[List['Synset']]:
        paths = list(self.relation_paths('hypernym', 'instance_hypernym'))
        if include_self:
            paths = [[self] + path for path in paths] or [[self]]
        if simulate_root and self.id != _FAKE_ROOT:
            root = Synset.empty(
                id=_FAKE_ROOT, _lexid=self._lexid, _wordnet=self._wordnet
            )
            paths = [path + [root] for path in paths] or [[root]]
        return paths

    def hypernym_paths(self, simulate_root: bool = False) -> List[List['Synset']]:
        """Return the list of hypernym paths to a root synset.

        Example:

            >>> for path in wn.synsets('dog', pos='n')[0].hypernym_paths():
            ...     for i, ss in enumerate(path):
            ...         print(' ' * i, ss, ss.lemmas()[0])
            ...
             Synset('pwn-02083346-n') canine
              Synset('pwn-02075296-n') carnivore
               Synset('pwn-01886756-n') eutherian mammal
                Synset('pwn-01861778-n') mammalian
                 Synset('pwn-01471682-n') craniate
                  Synset('pwn-01466257-n') chordate
                   Synset('pwn-00015388-n') animal
                    Synset('pwn-00004475-n') organism
                     Synset('pwn-00004258-n') animate thing
                      Synset('pwn-00003553-n') unit
                       Synset('pwn-00002684-n') object
                        Synset('pwn-00001930-n') physical entity
                         Synset('pwn-00001740-n') entity
             Synset('pwn-01317541-n') domesticated animal
              Synset('pwn-00015388-n') animal
               Synset('pwn-00004475-n') organism
                Synset('pwn-00004258-n') animate thing
                 Synset('pwn-00003553-n') unit
                  Synset('pwn-00002684-n') object
                   Synset('pwn-00001930-n') physical entity
                    Synset('pwn-00001740-n') entity

        """
        return self._hypernym_paths(simulate_root, False)

    def min_depth(self, simulate_root: bool = False) -> int:
        """Return the minimum taxonomy depth of the synset.

        Example:

            >>> wn.synsets('dog', pos='n')[0].min_depth()
            8

        """
        return min(
            (len(path) for path in self.hypernym_paths(simulate_root=simulate_root)),
            default=0
        )

    def max_depth(self, simulate_root: bool = False) -> int:
        """Return the maximum taxonomy depth of the synset.

        Example:

            >>> wn.synsets('dog', pos='n')[0].max_depth()
            13

        """
        return max(
            (len(path) for path in self.hypernym_paths(simulate_root=simulate_root)),
            default=0
        )

    def _shortest_hyp_paths(
            self, other: 'Synset', simulate_root: bool
    ) -> Dict[Tuple['Synset', int], List['Synset']]:
        if self == other:
            return {(self, 0): []}

        from_self = self._hypernym_paths(simulate_root, True)
        from_other = other._hypernym_paths(simulate_root, True)
        common = set(flatten(from_self)).intersection(flatten(from_other))

        if not common:
            return {}

        # Compute depths of common hypernyms from their distances.
        # Doing this now avoid more expensive lookups later.
        depths: Dict['Synset', int] = {}
        # subpaths accumulates paths to common hypernyms from both sides
        subpaths: Dict['Synset', Tuple[List[List['Synset']], List[List['Synset']]]]
        subpaths = {ss: ([], []) for ss in common}
        for which, paths in (0, from_self), (1, from_other):
            for path in paths:
                for dist, ss in enumerate(path):
                    if ss in common:
                        # self or other subpath to ss (not including ss)
                        subpaths[ss][which].append(path[:dist + 1])
                        # keep maximum depth
                        depth = len(path) - dist - 1
                        if ss not in depths or depths[ss] < depth:
                            depths[ss] = depth

        shortest: Dict[Tuple['Synset', int], List['Synset']] = {}
        for ss in common:
            from_self_subpaths, from_other_subpaths = subpaths[ss]
            shortest_from_self = min(from_self_subpaths, key=len)
            # for the other path, we need to reverse it and remove the pivot synset
            shortest_from_other = min(from_other_subpaths, key=len)[-2::-1]
            shortest[(ss, depths[ss])] = shortest_from_self + shortest_from_other

        return shortest

    def shortest_path(
            self, other: 'Synset', simulate_root: bool = False
    ) -> List['Synset']:
        """Return the shortest path from the synset to the *other* synset.

        Arguments:
            other: endpoint synset of the path
            simulate_root: if :python:`True`, ensure any two synsets
              are always connected by positing a fake root node

        """
        pathmap = self._shortest_hyp_paths(other, simulate_root)
        key = min(pathmap, key=lambda key: len(pathmap[key]), default=None)
        if key is None:
            raise wn.Error(f'no path between {self!r} and {other!r}')
        return pathmap[key][1:]

    def common_hypernyms(
            self, other: 'Synset', simulate_root: bool = False
    ) -> List['Synset']:
        """Return the common hypernyms for the current and *other* synsets.

        Arguments:
            other: synset that is a hyponym of any shared hypernyms
            simulate_root: if :python:`True`, ensure any two synsets
              always share a hypernym by positing a fake root node

        """
        from_self = self._hypernym_paths(simulate_root, True)
        from_other = other._hypernym_paths(simulate_root, True)
        common = set(flatten(from_self)).intersection(flatten(from_other))
        return sorted(common)

    def lowest_common_hypernyms(
            self, other: 'Synset', simulate_root: bool = False
    ) -> List['Synset']:
        """Return the common hypernyms furthest from the root.

        Arguments:
            other: synset that is a hyponym of any shared hypernyms
            simulate_root: if :python:`True`, ensure any two synsets
              always share a hypernym by positing a fake root node

        """
        pathmap = self._shortest_hyp_paths(other, simulate_root)
        # keys of pathmap are (synset, depth_of_synset)
        max_depth: int = max([depth for _, depth in pathmap], default=-1)
        if max_depth == -1:
            return []
        else:
            return [ss for ss, d in pathmap if d == max_depth]

    def holonyms(self) -> List['Synset']:
        """Return the list of synsets related by any holonym relation.

        Any of the following relations are traversed: ``holonym``,
        ``holo_location``, ``holo_member``, ``holo_part``,
        ``holo_portion``, ``holo_substance``.

        """
        return self.get_related(
            'holonym',
            'holo_location',
            'holo_member',
            'holo_part',
            'holo_portion',
            'holo_substance',
        )

    def meronyms(self) -> List['Synset']:
        """Return the list of synsets related by any meronym relation.

        Any of the following relations are traversed: ``meronym``,
        ``mero_location``, ``mero_member``, ``mero_part``,
        ``mero_portion``, ``mero_substance``.

        """
        return self.get_related(
            'meronym',
            'mero_location',
            'mero_member',
            'mero_part',
            'mero_portion',
            'mero_substance',
        )

    def hypernyms(self) -> List['Synset']:
        """Return the list of synsets related by any hypernym relation.

        Both the ``hypernym`` and ``instance_hypernym`` relations are
        traversed.

        """
        return self.get_related(
            'hypernym',
            'instance_hypernym'
        )

    def hyponyms(self) -> List['Synset']:
        """Return the list of synsets related by any hyponym relation.

        Both the ``hyponym`` and ``instance_hyponym`` relations are
        traversed.

        """
        return self.get_related(
            'hyponym',
            'instance_hyponym'
        )

    def translate(self, lexicon: str = None, *, lang: str = None) -> List['Synset']:
        """Return a list of translated synsets.

        Arguments:
            lexicon: if specified, translate to synsets in the target lexicon(s)
            lang: if specified, translate to synsets with the language code

        Example:

            >>> es = wn.synsets('araña', lang='es')[0]
            >>> en = es.translate(lexicon='ewn')[0]
            >>> en.lemmas()
            ['spider']

        """

        ili = self.ili
        if not ili:
            return []
        return synsets(ili=ili, lang=lang, lexicon=lexicon)


class Sense(_Relatable):
    """Class for modeling wordnet senses."""
    __slots__ = '_entry_id', '_synset_id'
    __module__ = 'wn'

    _ENTITY_TYPE = 'senses'

    def __init__(
            self,
            id: str,
            entry_id: str,
            synset_id: str,
            _lexid: int = NON_ROWID,
            _id: int = NON_ROWID,
            _wordnet: 'Wordnet' = None
    ):
        super().__init__(id=id, _lexid=_lexid, _id=_id, _wordnet=_wordnet)
        self._entry_id = entry_id
        self._synset_id = synset_id

    def __repr__(self) -> str:
        return f'Sense({self.id!r})'

    def word(self) -> Word:
        """Return the word of the sense.

        Example:

            >>> wn.senses('spigot')[0].word()
            Word('pwn-spigot-n')

        """
        return word(id=self._entry_id)

    def synset(self) -> Synset:
        """Return the synset of the sense.

        Example:

            >>> wn.senses('spigot')[0].synset()
            Synset('pwn-03325088-n')

        """
        return synset(id=self._synset_id)

    def examples(self) -> List[str]:
        """Return the list of examples for the sense."""
        exs = get_examples(self._id, 'senses')
        return [ex for ex, _, _ in exs]

    def lexicalized(self) -> bool:
        return get_lexicalized(self._id, 'senses')

    def metadata(self) -> Metadata:
        """Return the sense's metadata."""
        return get_metadata(self._id, 'senses')

    def get_related(self, *args: str) -> List['Sense']:
        """Return a list of related senses.

        One or more relation types should be passed as arguments which
        determine the kind of relations returned.

        Example:

            >>> physics = wn.senses('physics', lexicon='ewn')[0]
            >>> for sense in physics.get_related('has_domain_topic'):
            ...     print(sense.word().lemma())
            ...
            coherent
            chaotic
            incoherent

        """
        iterable = get_sense_relations(self._id, args)
        return [Sense(sid, eid, ssid, lexid, rowid, self._wordnet)
                for _, _, sid, eid, ssid, lexid, rowid in iterable]

    def get_related_synsets(self, *args: str) -> List[Synset]:
        """Return a list of related synsets."""
        iterable = get_sense_synset_relations(self._id, args)
        return [Synset(ssid, pos, ili, lexid, rowid, self._wordnet)
                for _, _, ssid, pos, ili, lexid, rowid in iterable]

    def translate(self, lexicon: str = None, *, lang: str = None) -> List['Sense']:
        """Return a list of translated senses.

        Arguments:
            lexicon: if specified, translate to senses in the target lexicon(s)
            lang: if specified, translate to senses with the language code

        Example:

            >>> en = wn.senses('petiole', lang='en')[0]
            >>> pt = en.translate(lang='pt')[0]
            >>> pt.word().lemma()
            'pecíolo'

        """
        synset = self.synset()
        return [t_sense
                for t_synset in synset.translate(lang=lang, lexicon=lexicon)
                for t_sense in t_synset.senses()]


class Wordnet:
    """Class for interacting with wordnet data.

    A wordnet object acts essentially as a filter by first selecting
    matching lexicons and then searching only within those lexicons
    for later queries. On instantiation, a *lang* argument is a BCP47
    language code that restricts the selected lexicons to those whose
    language matches the given code. A *lexicon* argument is a
    space-separated list of lexicon specifiers that more directly
    select lexicons by their ID and version; this is preferable when
    there are multiple lexicons in the same language or multiple
    version with the same ID.

    Some wordnets were created by translating the words from a larger
    wordnet, namely the Princeton WordNet, and then relying on the
    larger wordnet for structural relations. An *expand* argument is a
    second space-separated list of lexicon specifiers which are used
    for traversing relations, but not as the results of queries.

    """

    __slots__ = ('_lexicons', '_lexicon_ids', '_expanded', '_expanded_ids',
                 '_default_mode')
    __module__ = 'wn'

    def __init__(self, lexicon: str = None, *, lang: str = None, expand: str = None):
        # default mode means any lexicon is searched or expanded upon,
        # but relation traversals only target the source's lexicon
        self._default_mode = (not lexicon and not lang)

        lexs = list(find_lexicons(lexicon or '*', lang=lang))
        self._lexicons: Tuple[Lexicon, ...] = tuple(map(_to_lexicon, lexs))
        self._lexicon_ids: Tuple[int, ...] = tuple(lx._id for lx in self._lexicons)

        self._expanded: Tuple[Lexicon, ...] = ()
        if expand is None:
            if self._default_mode:
                expand = '*'
            # TODO: use project-specific settings
        if expand:
            self._expanded = tuple(map(_to_lexicon, find_lexicons(lexicon=expand)))
        self._expanded_ids: Tuple[int, ...] = tuple(lx._id for lx in self._expanded)

    def lexicons(self):
        """Return the list of lexicons covered by this wordnet."""
        return self._lexicons

    def expanded_lexicons(self):
        """Return the list of expand lexicons for this wordnet."""
        return self._expanded

    def word(self, id: str) -> Word:
        """Return the first word in this wordnet with identifier *id*."""
        iterable = find_entries(id=id, lexicon_rowids=self._lexicon_ids)
        try:
            return Word(*next(iterable), self)
        except StopIteration:
            raise wn.Error(f'no such lexical entry: {id}')

    def words(self, form: str = None, pos: str = None) -> List[Word]:
        """Return the list of matching words in this wordnet.

        Without any arguments, this function returns all words in the
        wordnet's selected lexicons. A *form* argument restricts the
        words to those matching the given word form, and *pos*
        restricts words by their part of speech.

        """
        iterable = find_entries(form=form, pos=pos, lexicon_rowids=self._lexicon_ids)
        return [Word(*word_data, self) for word_data in iterable]

    def synset(self, id: str) -> Synset:
        """Return the first synset in this wordnet with identifier *id*."""
        iterable = find_synsets(id=id, lexicon_rowids=self._lexicon_ids)
        try:
            return Synset(*next(iterable), self)
        except StopIteration:
            raise wn.Error(f'no such synset: {id}')

    def synsets(
        self, form: str = None, pos: str = None, ili: str = None
    ) -> List[Synset]:
        """Return the list of matching synsets in this wordnet.

        Without any arguments, this function returns all synsets in
        the wordnet's selected lexicons. A *form* argument restricts
        synsets to those whose member words match the given word
        form. A *pos* argument restricts synsets to those with the
        given part of speech. An *ili* argument restricts synsets to
        those with the given interlingual index; generally this should
        select a unique synset within a single lexicon.

        """
        iterable = find_synsets(
            form=form, pos=pos, ili=ili, lexicon_rowids=self._lexicon_ids,
        )
        return [Synset(*synset_data, self) for synset_data in iterable]

    def sense(self, id: str) -> Sense:
        """Return the first sense in this wordnet with identifier *id*."""
        iterable = find_senses(id=id, lexicon_rowids=self._lexicon_ids)
        try:
            return Sense(*next(iterable), self)
        except StopIteration:
            raise wn.Error(f'no such sense: {id}')

    def senses(self, form: str = None, pos: str = None) -> List[Sense]:
        """Return the list of matching senses in this wordnet.

        Without any arguments, this function returns all senses in the
        wordnet's selected lexicons. A *form* argument restricts the
        senses to those whose word matches the given word form, and
        *pos* restricts senses by their word's part of speech.

        """
        iterable = find_senses(form=form, pos=pos, lexicon_rowids=self._lexicon_ids)
        return [Sense(*sense_data, self) for sense_data in iterable]


def _to_lexicon(data) -> Lexicon:
    rowid, id, label, language, email, license, version, url, citation = data
    return Lexicon(
        id,
        label,
        language,
        email,
        license,
        version,
        url=url,
        citation=citation,
        _id=rowid
    )


def projects() -> List[Dict]:
    """Return the list of indexed projects.

    This returns the same dictionaries of information as
    :meth:`wn.config.get_project_info
    <wn._config.WNConfig.get_project_info>`, but for all indexed
    projects.

    Example:

        >>> infos = wn.projects()
        >>> len(infos)
        36
        >>> infos[0]['label']
        'Open English WordNet'

    """
    index = wn.config.index
    return [
        wn.config.get_project_info(f'{project_id}:{version}')
        for project_id, project_info in index.items()
        for version in project_info['versions']
    ]


def lexicons(*, lexicon: str = None, lang: str = None) -> List[Lexicon]:
    """Return the lexicons matching a language or lexicon specifier.

    Example:

        >>> wn.lexicons(lang='en')
        [<Lexicon ewn:2020 [en]>, <Lexicon pwn:3.0 [en]>]

    """
    if lexicon is None:
        lexicon = '*'
    try:
        w = Wordnet(lang=lang, lexicon=lexicon)
    except wn.Error:
        return []
    else:
        return w.lexicons()


def word(id: str, *, lexicon: str = None, lang: str = None) -> Word:
    """Return the word with *id* in *lexicon*.

    This will create a :class:`Wordnet` object using the *lang* and
    *lexicon* arguments. The *id* argument is then passed to the
    :meth:`Wordnet.word` method.

    >>> wn.word('ewn-cell-n')
    Word('ewn-cell-n')

    """
    return Wordnet(lang=lang, lexicon=lexicon).word(id=id)


def words(
    form: str = None,
    pos: str = None,
    *,
    lexicon: str = None,
    lang: str = None,
) -> List[Word]:
    """Return the list of matching words.

    This will create a :class:`Wordnet` object using the *lang* and
    *lexicon* arguments. The remaining arguments are passed to the
    :meth:`Wordnet.words` method.

    >>> len(wn.words())
    282902
    >>> len(wn.words(pos='v'))
    34592
    >>> wn.words(form="scurry")
    [Word('ewn-scurry-n'), Word('ewn-scurry-v')]

    """
    return Wordnet(lang=lang, lexicon=lexicon).words(form=form, pos=pos)


def synset(id: str, *, lexicon: str = None, lang: str = None) -> Synset:
    """Return the synset with *id* in *lexicon*.

    This will create a :class:`Wordnet` object using the *lang* and
    *lexicon* arguments. The *id* argument is then passed to the
    :meth:`Wordnet.synset` method.

    >>> wn.synset('ewn-03311152-n')
    Synset('ewn-03311152-n')

    """
    return Wordnet(lang=lang, lexicon=lexicon).synset(id=id)


def synsets(
    form: str = None,
    pos: str = None,
    ili: str = None,
    *,
    lexicon: str = None,
    lang: str = None,
) -> List[Synset]:
    """Return the list of matching synsets.

    This will create a :class:`Wordnet` object using the *lang* and
    *lexicon* arguments. The remaining arguments are passed to the
    :meth:`Wordnet.synsets` method.

    >>> len(wn.synsets('couch'))
    4
    >>> wn.synsets('couch', pos='v')
    [Synset('ewn-00983308-v')]

    """
    return Wordnet(lang=lang, lexicon=lexicon).synsets(form=form, pos=pos, ili=ili)


def senses(
    form: str = None,
    pos: str = None,
    *,
    lexicon: str = None,
    lang: str = None,
) -> List[Sense]:
    """Return the list of matching senses.

    This will create a :class:`Wordnet` object using the *lang* and
    *lexicon* arguments. The remaining arguments are passed to the
    :meth:`Wordnet.senses` method.

    >>> len(wn.senses('twig'))
    3
    >>> wn.senses('twig', pos='n')
    [Sense('ewn-twig-n-13184889-02')]

    """
    return Wordnet(lang=lang, lexicon=lexicon).senses(form=form, pos=pos)


def sense(id: str, *, lexicon: str = None, lang: str = None) -> Sense:
    """Return the sense with *id* in *lexicon*.

    This will create a :class:`Wordnet` object using the *lang* and
    *lexicon* arguments. The *id* argument is then passed to the
    :meth:`Wordnet.sense` method.

    >>> wn.sense('ewn-flutter-v-01903884-02')
    Sense('ewn-flutter-v-01903884-02')

    """
    return Wordnet(lang=lang, lexicon=lexicon).sense(id=id)




"""
Local configuration settings.
"""

from typing import Dict
from pathlib import Path

import toml

from wn import Error
from wn._types import AnyPath
from wn._util import is_url, resources, short_hash

# The directory where downloaded and added data will be stored.
DEFAULT_DATA_DIRECTORY = Path.home() / '.wn_data'
DATABASE_FILENAME = 'wn.db'


class WNConfig:

    def __init__(self):
        self._data_directory = DEFAULT_DATA_DIRECTORY
        self._projects = {}
        self._dbpath = self._data_directory / DATABASE_FILENAME
        self.allow_multithreading = False

    @property
    def data_directory(self) -> Path:
        """The file system directory where Wn's data is stored."""
        dir = self._data_directory
        dir.mkdir(exist_ok=True)
        return dir

    @data_directory.setter
    def data_directory(self, path):
        dir = Path(path).expanduser()
        if dir.exists() and not dir.is_dir():
            raise Error(f'path exists and is not a directory: {dir}')
        self._data_directory = dir
        self._dbpath = dir / DATABASE_FILENAME

    @property
    def database_path(self):
        """The path to the database file."""
        return self._dbpath

    @property
    def downloads_directory(self):
        """The file system directory where downloads are cached."""
        dir = self.data_directory / 'downloads'
        dir.mkdir(exist_ok=True)
        return dir

    @property
    def index(self) -> Dict[str, Dict]:
        """The project index."""
        return self._projects

    def add_project(
            self,
            id: str,
            label: str,
            language: str,
            license: str = None,
    ) -> None:
        """Add a new wordnet project to the index.

        Arguments:
            id: short identifier of the project
            label: full name of the project
            language: `BCP 47`_ language code of the resource
            license: link or name of the project's default license

        .. _BCP 47: https://en.wikipedia.org/wiki/IETF_language_tag
        """
        if id in self._projects:
            raise ValueError(f'project already added: {id}')
        self._projects[id] = {
            'label': label,
            'language': language,
            'versions': {},
            'license': license
        }

    def add_project_version(
            self,
            id: str,
            version: str,
            url: str,
            license: str = None,
    ) -> None:
        """Add a new resource version for a project.

        Arguments:
            id: short identifier of the project
            version: version string of the resource
            url: web address of the resource
            license: link or name of the resource's license; if not
              given, the project's default license will be used.

        """
        version_data = {'resource_url': url}
        if license:
            version_data['license'] = license
        project = self._projects[id]
        project['versions'][version] = version_data

    def get_project_info(self, arg: str) -> Dict:
        """Return a dictionary of information about an indexed project.

        If the project has been downloaded and cached, the ``"cache"``
        key will point to the path of the cached file, otherwise its
        value is ``None``.

        Arguments:
            arg: a project specifier

        Example:

            >>> info = wn.config.get_project_info('pwn:3.0')
            >>> info['label']
            'Princeton WordNet'

        """
        id, _, version = arg.partition(':')
        project: Dict = self._projects[id]
        versions: Dict = project['versions']
        if not version or version == '*':
            version = next(iter(versions))
        if version not in versions:
            raise Error(f'no such version: {version!r} ({project})')

        url = versions[version]['resource_url']
        cache_path = self.get_cache_path(url)

        return dict(
            id=id,
            version=version,
            label=project['label'],
            language=project['language'],
            license=versions[version].get('license', project.get('license')),
            resource_url=url,
            cache=cache_path if cache_path.exists() else None
        )

    def get_cache_path(self, arg: str) -> Path:
        """Return the path for caching *arg*.

        The *arg* argument may be either a URL or a project specifier
        that gets passed to :meth:`get_project_info`. Note that this
        is just a path operation and does not signify that the file
        exists in the file system.

        """
        if not is_url(arg):
            arg = self.get_project_info(arg)['resource_url']
        filename = short_hash(arg)
        return self.downloads_directory / filename

    def update(self, data: dict) -> None:
        """Update the configuration with items in *data*.

        Items are only inserted or replaced, not deleted. If a project
        index is provided in the ``"index"`` key, then either the
        project must not already be indexed or any project fields
        (label, language, or license) that are specified must be equal
        to the indexed project.

        """
        if 'data_directory' in data:
            self.data_directory = data['data_directory']
        for id, project in data.get('index', {}).items():
            if id in self._projects:
                # validate that they are the same
                _project = self._projects[id]
                for attr in ('label', 'language', 'license'):
                    if attr in project and project[attr] != _project[attr]:
                        raise Error(f'{attr} mismatch for {id}')
            else:
                self.add_project(
                    id,
                    project['label'],
                    project['language'],
                    license=project.get('license'),
                )
            for version, info in project.get('versions', {}).items():
                self.add_project_version(
                    id,
                    version,
                    info['url'],
                    license=info.get('license'),
                )

    def load_index(self, path: AnyPath) -> None:
        """Load and update with the project index at *path*.

        The project index is a TOML_ file containing project and
        version information. For example:

        .. code-block:: toml

           [ewn]
             label = "Open English WordNet"
             language = "en"
             license = "https://creativecommons.org/licenses/by/4.0/"
             [ewn.versions.2019]
               url = "https://en-word.net/static/english-wordnet-2019.xml.gz"
             [ewn.versions.2020]
               url = "https://en-word.net/static/english-wordnet-2020.xml.gz"

        .. _TOML: https://toml.io

        """
        path = Path(path).expanduser()
        index = toml.load(path)
        self.update({'index': index})


config = WNConfig()
with resources.path('wn', 'index.toml') as index_path:
    config.load_index(index_path)




import pytest

import wn


@pytest.mark.usefixtures('mini_db')
def test_hypernym_paths():
    information = wn.synsets('information')[0]
    example = wn.synsets('example')[0]
    sample = wn.synsets('sample')[0]
    random_sample = wn.synsets('random sample')[0]
    assert information.hypernym_paths() == []
    assert example.hypernym_paths() == [[information]]
    assert sample.hypernym_paths() == [[example, information]]
    assert random_sample.hypernym_paths() == [[sample, example, information]]


@pytest.mark.usefixtures('mini_db')
def test_interlingual_hypernym_paths():
    información = wn.synsets('información')[0]
    ejemplo = wn.synsets('ejemplo')[0]
    inferred = wn.Synset.empty('*INFERRED*')
    muestra_aleatoria = wn.synsets('muestra aleatoria')[0]
    assert información.hypernym_paths() == []
    assert ejemplo.hypernym_paths() == [[información]]
    assert muestra_aleatoria.hypernym_paths() == [[inferred, ejemplo, información]]


@pytest.mark.usefixtures('mini_db')
def test_shortest_path():
    information = wn.synsets('information')[0]
    example = wn.synsets('example')[0]
    sample = wn.synsets('sample')[0]
    random_sample = wn.synsets('random sample')[0]
    datum = wn.synsets('datum')[0]
    exemplify = wn.synsets('exemplify')[0]
    inferred_root = wn.Synset.empty('*INFERRED*')
    assert information.shortest_path(information) == []
    assert information.shortest_path(datum) == [datum]
    assert information.shortest_path(sample) == [example, sample]
    assert sample.shortest_path(information) == [example, information]
    assert random_sample.shortest_path(datum) == [sample, example, information, datum]
    with pytest.raises(wn.Error):
        example.shortest_path(exemplify)
    assert example.shortest_path(exemplify, simulate_root=True) == [
        information, inferred_root, exemplify
    ]


@pytest.mark.usefixtures('mini_db')
def test_min_depth():
    assert wn.synsets('information')[0].min_depth() == 0
    assert wn.synsets('example')[0].min_depth() == 1
    assert wn.synsets('sample')[0].min_depth() == 2
    assert wn.synsets('random sample')[0].min_depth() == 3


@pytest.mark.usefixtures('mini_db')
def test_max_depth():
    assert wn.synsets('information')[0].max_depth() == 0
    assert wn.synsets('example')[0].max_depth() == 1
    assert wn.synsets('sample')[0].max_depth() == 2
    assert wn.synsets('random sample')[0].max_depth() == 3




"""
Reader for the Lexical Markup Framework (LMF) format.
"""

from typing import (
    TypeVar,
    Type,
    Container,
    List,
    Tuple,
    Dict,
    Set,
    NamedTuple,
    Optional,
    TextIO,
)
from pathlib import Path
import warnings
import xml.etree.ElementTree as ET  # for general XML parsing
import xml.parsers.expat  # for fast scanning of Lexicon versions
from xml.sax.saxutils import quoteattr

import wn
from wn._types import AnyPath
from wn._util import is_xml
from wn.constants import (
    SENSE_RELATIONS,
    SYNSET_RELATIONS,
    ADJPOSITIONS,
    PARTS_OF_SPEECH,
)


class LMFError(wn.Error):
    """Raised on invalid LMF-XML documents."""


class LMFWarning(Warning):
    """Issued on non-conforming LFM values."""


_XMLDECL = '<?xml version="1.0" encoding="UTF-8"?>'
_DOCTYPE = '<!DOCTYPE LexicalResource SYSTEM "{schema}">'
_SCHEMAS = {
    '1.0': 'http://globalwordnet.github.io/schemas/WN-LMF-1.0.dtd',
    '1.1': 'http://globalwordnet.github.io/schemas/WN-LMF-1.1.dtd',
}
_DOCTYPES = {
    _DOCTYPE.format(schema=schema): version for version, schema in _SCHEMAS.items()
}

_DC_URIS = {
    '1.0': 'http://purl.org/dc/elements/1.1/',
    '1.1': 'http://globalwordnet.github.io/schemas/dc/',
}
_DC_ATTRS = [
    'contributor',
    'coverage',
    'creator',
    'date',
    'description',
    'format',
    'identifier',
    'publisher',
    'relation',
    'rights',
    'source',
    'subject',
    'title',
    'type',
]
# dublin-core metadata needs the namespace uri prefixed
_DC_QNAME_PAIRS = {
    version: [(f'{{{uri}}}{attr}', attr) for attr in _DC_ATTRS]
    for version, uri in _DC_URIS.items()
}


class Metadata(NamedTuple):
    contributor: Optional[str] = None
    coverage: Optional[str] = None
    creator: Optional[str] = None
    date: Optional[str] = None
    description: Optional[str] = None
    format: Optional[str] = None
    identifier: Optional[str] = None
    publisher: Optional[str] = None
    relation: Optional[str] = None
    rights: Optional[str] = None
    source: Optional[str] = None
    subject: Optional[str] = None
    title: Optional[str] = None
    type: Optional[str] = None
    status: Optional[str] = None
    note: Optional[str] = None
    confidence: Optional[float] = None


class _HasMeta:
    __slots__ = 'meta',

    def __init__(self, meta: Metadata = None):
        self.meta = meta


# These types model the WN-LMF DTD
# http://globalwordnet.github.io/schemas/WN-LMF-1.0.dtd

class Count(_HasMeta):
    __slots__ = 'value',

    def __init__(self, value: int, meta: Metadata = None):
        super().__init__(meta)
        self.value = value


class SyntacticBehaviour:
    __slots__ = 'frame', 'senses'

    def __init__(self, frame: str, senses: List[str] = None):
        self.frame = frame
        self.senses = senses or []


class SenseRelation(_HasMeta):
    __slots__ = 'target', 'type'

    def __init__(self, target: str, type: str, meta: Metadata = None):
        super().__init__(meta)
        self.target = target
        self.type = type


class SynsetRelation(_HasMeta):
    __slots__ = 'target', 'type'

    def __init__(self, target: str, type: str, meta: Metadata = None):
        super().__init__(meta)
        self.target = target
        self.type = type


class Example(_HasMeta):
    __slots__ = 'text', 'language'

    def __init__(self, text: str, language: str = '', meta: Metadata = None):
        super().__init__(meta)
        self.text = text
        self.language = language


class ILIDefinition(_HasMeta):
    __slots__ = 'text',

    def __init__(self, text: str, meta: Metadata = None):
        super().__init__(meta)
        self.text = text


class Definition(_HasMeta):
    __slots__ = 'text', 'language', 'source_sense'

    def __init__(
            self,
            text: str,
            language: str = '',
            source_sense: str = '',
            meta: Metadata = None):
        super().__init__(meta)
        self.text = text
        self.language = language
        self.source_sense = source_sense


class Synset(_HasMeta):
    __slots__ = ('id', 'ili', 'pos', 'definitions', 'ili_definition',
                 'relations', 'examples', 'lexicalized')

    def __init__(
            self,
            id: str,
            ili: str,
            pos: str,
            definitions: List[Definition] = None,
            ili_definition: ILIDefinition = None,
            relations: List[SynsetRelation] = None,
            examples: List[Example] = None,
            lexicalized: bool = True,
            meta: Metadata = None):
        super().__init__(meta)
        self.id = id
        self.ili = ili
        self.pos = pos
        self.definitions = definitions or []
        self.ili_definition = ili_definition
        self.relations = relations or []
        self.examples = examples or []
        self.lexicalized = lexicalized


class Sense(_HasMeta):
    __slots__ = ('id', 'synset', 'relations', 'examples', 'counts',
                 'lexicalized', 'adjposition')

    def __init__(
            self,
            id: str,
            synset: str,
            relations: List[SenseRelation] = None,
            examples: List[Example] = None,
            counts: List[Count] = None,
            lexicalized: bool = True,
            adjposition: str = '',
            meta: Metadata = None):
        super().__init__(meta)
        self.id = id
        self.synset = synset
        self.relations = relations or []
        self.examples = examples or []
        self.counts = counts or []
        self.lexicalized = lexicalized
        self.adjposition = adjposition


class Tag:
    __slots__ = 'text', 'category'

    def __init__(self, text: str, category: str):
        self.text = text
        self.category = category


class Form:
    __slots__ = 'form', 'script', 'tags'

    def __init__(self, form: str, script: str, tags: List[Tag] = None):
        self.form = form
        self.script = script
        self.tags = tags or []


class Lemma:
    __slots__ = 'form', 'pos', 'script', 'tags'

    def __init__(self, form: str, pos: str, script: str = '', tags: List[Tag] = None):
        self.form = form
        self.pos = pos
        self.script = script
        self.tags = tags or []


class LexicalEntry(_HasMeta):
    __slots__ = 'id', 'lemma', 'forms', 'senses'

    def __init__(
            self,
            id: str,
            lemma: Lemma,
            forms: List[Form] = None,
            senses: List[Sense] = None,
            meta: Metadata = None):
        super().__init__(meta)
        self.id = id
        self.lemma = lemma
        self.forms = forms or []
        self.senses = senses or []


class Lexicon(_HasMeta):
    __slots__ = ('id', 'label', 'language',
                 'email', 'license', 'version', 'url', 'citation',
                 'lexical_entries', 'synsets', 'syntactic_behaviours')

    def __init__(
            self,
            id: str,
            label: str,
            language: str,
            email: str,
            license: str,
            version: str,
            url: str = '',
            citation: str = '',
            lexical_entries: List[LexicalEntry] = None,
            synsets: List[Synset] = None,
            syntactic_behaviours: List[SyntacticBehaviour] = None,
            meta: Metadata = None):
        super().__init__(meta)
        self.id = id
        self.label = label
        self.language = language
        self.email = email
        self.license = license
        self.version = version
        self.url = url
        self.citation = citation
        self.lexical_entries = lexical_entries or []
        self.synsets = synsets or []
        self.syntactic_behaviours = syntactic_behaviours or []

    def entry_ids(self) -> Set[str]:
        return {entry.id for entry in self.lexical_entries}

    def sense_ids(self) -> Set[str]:
        return {sense.id for entry in self.lexical_entries for sense in entry.senses}

    def synset_ids(self) -> Set[str]:
        return {synset.id for synset in self.synsets}


LexicalResource = List[Lexicon]


def is_lmf(source: AnyPath) -> bool:
    """Return True if *source* is a WN-LMF file."""
    source = Path(source).expanduser()
    if not is_xml(source):
        return False
    with source.open() as fh:
        xmldecl, doctype = _read_header(fh)
        if not (xmldecl == _XMLDECL and doctype in _DOCTYPES):
            return False
    return True


def _read_header(fh: TextIO) -> Tuple[str, str]:
    return (fh.readline().rstrip().replace("'", '"'),  # XML declaration
            fh.readline().rstrip().replace("'", '"'))  # DOCTYPE declaration


def scan_lexicons(source: AnyPath) -> List[Dict]:
    """Scan *source* and return only the top-level lexicon info."""

    source = Path(source).expanduser()

    # this is implemeted with expat as it's much faster than etree for
    # this task
    infos = []

    def start(name, attrs):
        if name == 'Lexicon':
            attrs['counts'] = {}
            infos.append(attrs)
        elif infos:
            counts = infos[-1]['counts']
            counts[name] = counts.get(name, 0) + 1

    p = xml.parsers.expat.ParserCreate()
    p.StartElementHandler = start
    with open(source, 'rb') as fh:
        try:
            p.ParseFile(fh)
        except xml.parsers.expat.ExpatError as exc:
            raise LMFError('invalid or ill-formed WN-LMF file') from exc

    return infos


def load(source: AnyPath) -> LexicalResource:
    """Load wordnets encoded in the WN-LMF format.

    Args:
        source: path to a WN-LMF file
    """
    source = Path(source).expanduser()

    with source.open('rt') as fh:
        xmldecl, doctype = _read_header(fh)
        assert xmldecl == _XMLDECL
        version = _DOCTYPES[doctype]

    events = ET.iterparse(source, events=('start', 'end'))
    root = next(events)[1]
    event, elem = next(events)

    lexicons: List[Lexicon] = []
    while event == 'start' and elem.tag == 'Lexicon':
        lexicons.append(_load_lexicon(elem, events, version))
        root.clear()
        event, elem = next(events)

    _assert_closed(event, elem, 'LexicalResource')
    list(events)  # consume remaining events, if any

    return lexicons


def dump(
        lexicons: LexicalResource, destination: AnyPath, version: str = '1.0'
) -> None:
    """Write wordnets in the WN-LMF format.

    Args:
        lexicons: a list of :class:`Lexicon` objects
    """
    destination = Path(destination).expanduser()
    doctype = _DOCTYPE.format(schema=_SCHEMAS[version])
    dc_uri = _DC_URIS[version]
    with destination.open('wt', encoding='utf-8') as out:
        print(_XMLDECL, file=out)
        print(doctype, file=out)
        print(f'<LexicalResource xmlns:dc="{dc_uri}">', file=out)
        for lexicon in lexicons:
            _dump_lexicon(lexicon, out, version)
        print('</LexicalResource>', file=out)


def _load_lexicon(local_root, events, version) -> Lexicon:
    attrs = local_root.attrib
    event, elem = next(events)

    syntactic_behaviours: List[SyntacticBehaviour] = []
    lexical_entries: List[LexicalEntry] = []
    while event == 'start' and elem.tag == 'LexicalEntry':
        entry, sbs = _load_lexical_entry(elem, events, version)
        lexical_entries.append(entry)
        syntactic_behaviours.extend(sbs)
        local_root.clear()
        event, elem = next(events)

    synsets: List[Synset] = []
    while event == 'start' and elem.tag == 'Synset':
        synsets.append(_load_synset(elem, events, version))
        local_root.clear()
        event, elem = next(events)

    _assert_closed(event, elem, 'Lexicon')

    return Lexicon(
        attrs['id'],
        attrs['label'],
        attrs['language'],
        attrs['email'],
        attrs['license'],
        attrs['version'],
        url=attrs.get('url'),
        citation=attrs.get('citation'),
        lexical_entries=lexical_entries,
        synsets=synsets,
        syntactic_behaviours=syntactic_behaviours,
        meta=_get_metadata(attrs, version),
    )


def _load_lexical_entry(
        local_root,
        events,
        version,
) -> Tuple[LexicalEntry, List[SyntacticBehaviour]]:
    attrs = local_root.attrib
    lemma: Lemma = _load_lemma(events)
    event, elem = next(events)

    forms: List[Form] = []
    while event == 'start' and elem.tag == 'Form':
        forms.append(_load_form(elem, events))
        event, elem = next(events)

    senses: List[Sense] = []
    while event == 'start' and elem.tag == 'Sense':
        senses.append(_load_sense(elem, events, version))
        event, elem = next(events)

    syntactic_behaviours: List[SyntacticBehaviour] = []
    while event == 'start' and elem.tag == 'SyntacticBehaviour':
        event, elem = next(events)
        _assert_closed(event, elem, 'SyntacticBehaviour')
        syntactic_behaviours.append(_load_syntactic_behaviour(elem))
        event, elem = next(events)

    _assert_closed(event, elem, 'LexicalEntry')

    entry = LexicalEntry(
        attrs['id'],
        lemma,
        forms=forms,
        senses=senses,
        meta=_get_metadata(attrs, version),
    )

    return entry, syntactic_behaviours


def _load_lemma(events) -> Lemma:
    event, elem = next(events)
    if event != 'start' or elem.tag != 'Lemma':
        raise LMFError('expected a Lemma element')
    attrs = elem.attrib
    return Lemma(
        attrs['writtenForm'],
        _get_literal(attrs['partOfSpeech'], PARTS_OF_SPEECH),
        script=attrs.get('script'),
        tags=_load_tags_until(events, 'Lemma'))


def _load_form(local_root, events) -> Form:
    attrs = local_root.attrib
    return Form(
        attrs['writtenForm'],
        script=attrs.get('script'),
        tags=_load_tags_until(events, 'Form'))


def _load_tags_until(events, terminus) -> List[Tag]:
    event, elem = next(events)

    tags: List[Tag] = []
    while event == 'start' and elem.tag == 'Tag':
        event, elem = next(events)
        _assert_closed(event, elem, 'Tag')
        tags.append(Tag(elem.text, elem.attrib['category']))
        event, elem = next(events)

    _assert_closed(event, elem, terminus)

    return tags


def _load_sense(local_root, events, version) -> Sense:
    attrs = local_root.attrib
    event, elem = next(events)

    relations: List[SenseRelation] = []
    while event == 'start' and elem.tag == 'SenseRelation':
        event, elem = next(events)
        _assert_closed(event, elem, 'SenseRelation')
        relations.append(
            _load_relation(elem, SenseRelation, SENSE_RELATIONS, version)
        )
        event, elem = next(events)

    examples: List[Example] = []
    while event == 'start' and elem.tag == 'Example':
        event, elem = next(events)
        _assert_closed(event, elem, 'Example')
        examples.append(_load_example(elem, version))
        event, elem = next(events)

    counts: List[Count] = []
    while event == 'start' and elem.tag == 'Count':
        event, elem = next(events)
        _assert_closed(event, elem, 'Count')
        counts.append(_load_count(elem, version))
        event, elem = next(events)

    _assert_closed(event, elem, 'Sense')

    return Sense(
        attrs['id'],
        attrs['synset'],
        relations=relations,
        examples=examples,
        counts=counts,
        lexicalized=_get_bool(attrs.get('lexicalized', 'true')),
        adjposition=_get_optional_literal(attrs.get('adjposition'), ADJPOSITIONS),
        meta=_get_metadata(attrs, version),
    )


_R = TypeVar('_R', SynsetRelation, SenseRelation)


def _load_relation(elem, cls: Type[_R], choices: Container[str], version: str) -> _R:
    attrs = elem.attrib
    return cls(
        attrs['target'],
        _get_literal
        (attrs['relType'], choices),
        meta=_get_metadata(attrs, version),
    )


def _load_example(elem, version) -> Example:
    attrs = elem.attrib
    return Example(
        elem.text,
        language=attrs.get('language'),
        meta=_get_metadata(attrs, version),
    )


def _load_count(elem, version) -> Count:
    attrs = elem.attrib
    value: int = -1
    try:
        value = int(elem.text)
    except ValueError:
        warnings.warn(f'count must be an integer: {elem.text}', LMFWarning)
    return Count(
        value,
        meta=_get_metadata(attrs, version),
    )


def _load_syntactic_behaviour(elem) -> SyntacticBehaviour:
    attrs = elem.attrib
    return SyntacticBehaviour(
        attrs['subcategorizationFrame'],
        senses=attrs.get('senses', '').split(),
    )


def _load_synset(local_root, events, version) -> Synset:
    attrs = local_root.attrib
    event, elem = next(events)

    definitions: List[Definition] = []
    while event == 'start' and elem.tag == 'Definition':
        event, elem = next(events)
        _assert_closed(event, elem, 'Definition')
        definitions.append(_load_definition(elem, version))
        event, elem = next(events)

    ili_definition = None
    if event == 'start' and elem.tag == 'ILIDefinition':
        event, elem = next(events)
        _assert_closed(event, elem, 'ILIDefinition')
        ili_definition = _load_ilidefinition(elem, version)
        event, elem = next(events)

    relations: List[SynsetRelation] = []
    while event == 'start' and elem.tag == 'SynsetRelation':
        event, elem = next(events)
        _assert_closed(event, elem, 'SynsetRelation')
        relations.append(
            _load_relation(elem, SynsetRelation, SYNSET_RELATIONS, version)
        )
        event, elem = next(events)

    examples: List[Example] = []
    while event == 'start' and elem.tag == 'Example':
        event, elem = next(events)
        _assert_closed(event, elem, 'Example')
        examples.append(_load_example(elem, version))
        event, elem = next(events)

    _assert_closed(event, elem, 'Synset')

    return Synset(
        attrs['id'],
        attrs['ili'],
        pos=_get_optional_literal(attrs.get('partOfSpeech'), PARTS_OF_SPEECH),
        definitions=definitions,
        ili_definition=ili_definition,
        relations=relations,
        examples=examples,
        lexicalized=_get_bool(attrs.get('lexicalized', 'true')),
        meta=_get_metadata(attrs, version),
    )


def _load_definition(elem, version) -> Definition:
    attrs = elem.attrib
    return Definition(
        elem.text,
        language=attrs.get('language'),
        source_sense=attrs.get('sourceSense'),
        meta=_get_metadata(attrs, version),
    )


def _load_ilidefinition(elem, version) -> ILIDefinition:
    attrs = elem.attrib
    return ILIDefinition(
        elem.text,
        meta=_get_metadata(attrs, version),
    )


def _get_bool(value: str) -> bool:
    value = value.lower()
    if value not in ('true', 'false'):
        warnings.warn(f'value must be "true" or "false", not {value!r}', LMFWarning)
    return value == 'true'


def _get_optional_literal(value: Optional[str], choices: Container[str]) -> str:
    if value is None:
        return ''
    else:
        return _get_literal(value, choices)


def _get_literal(value: str, choices: Container[str]) -> str:
    if value is not None and value not in choices:
        warnings.warn(f'{value!r} is not one of {choices!r}', LMFWarning)
    return value


def _get_metadata(attrs: Dict, version: str) -> Optional[Metadata]:
    metas = [attrs.get(qname) for qname, _ in _DC_QNAME_PAIRS[version]]
    metas.append(attrs.get('status'))
    metas.append(attrs.get('note'))
    if 'confidenceScore' in attrs:
        value = attrs['confidenceScore']
        try:
            value = float(value)
            assert 0 <= value <= 1
        except (ValueError, AssertionError):
            warnings.warn(f'confidenceScore not between 0 and 1: {value}', LMFWarning)
        metas.append(value)
    else:
        metas.append(None)
    if any(meta is not None for meta in metas):
        return Metadata(*metas)
    else:
        return None


def _assert_closed(event, elem, tag) -> None:
    if event != 'end' or elem.tag != tag:
        raise _unexpected(elem)


def _unexpected(elem) -> LMFError:
    return LMFError(f'unexpected element: {elem.tag}')


def _dump_lexicon(lexicon: Lexicon, out: TextIO, version: str) -> None:
    attrib = {
        'id': lexicon.id,
        'label': lexicon.label,
        'language': lexicon.language,
        'email': lexicon.email,
        'license': lexicon.license,
        'version': lexicon.version,
    }
    if lexicon.url:
        attrib['url'] = lexicon.url
    if lexicon.citation:
        attrib['citation'] = lexicon.citation
    attrib.update(_meta_dict(lexicon.meta))
    attrdelim = '\n' + (' ' * 11)
    attrs = attrdelim.join(f'{attr}={quoteattr(val)}' for attr, val in attrib.items())
    print(f'  <Lexicon {attrs}>', file=out)

    for entry in lexicon.lexical_entries:
        # TODO: 1.0 SyntacticBehaviour
        _dump_lexical_entry(entry, out, version)

    for synset in lexicon.synsets:
        _dump_synset(synset, out, version)

    # TODO: 1.1 SyntacticBehaviour

    print('  </Lexicon>', file=out)


def _dump_lexical_entry(
        entry: LexicalEntry, out: TextIO, version: str
) -> None:
    attrib = {'id': entry.id}
    attrib.update(_meta_dict(entry.meta))
    elem = ET.Element('LexicalEntry', attrib=attrib)
    elem.append(_build_lemma(entry.lemma))
    elem.extend(_build_form(form) for form in entry.forms)
    elem.extend(_build_sense(sense) for sense in entry.senses)
    # TODO: 1.0 SyntacticBehaviour
    print(_tostring(elem, 2), file=out)


def _build_lemma(lemma: Lemma) -> ET.Element:
    attrib = {'writtenForm': lemma.form}
    if lemma.script:
        attrib['script'] = lemma.script
    attrib['partOfSpeech'] = lemma.pos
    elem = ET.Element('Lemma', attrib=attrib)
    # TODO: Pronunciation
    for tag in lemma.tags:
        elem.append(_build_tag(tag))
    return elem


def _build_form(form: Form) -> ET.Element:
    attrib = {'writtenForm': form.form}
    if form.script:
        attrib['script'] = form.script
    elem = ET.Element('Form', attrib=attrib)
    # TODO: Pronunciation
    for tag in form.tags:
        elem.append(_build_tag(tag))
    return elem


def _build_tag(tag: Tag) -> ET.Element:
    elem = ET.Element('Tag', category=tag.category)
    elem.text = tag.text
    return elem


def _build_sense(sense: Sense) -> ET.Element:
    attrib = {'id': sense.id, 'synset': sense.synset}
    attrib.update(_meta_dict(sense.meta))
    if not sense.lexicalized:
        attrib['lexicalized'] = 'false'
    if sense.adjposition:
        attrib['adjposition'] = sense.adjposition
    # TODO: subcat
    elem = ET.Element('Sense', attrib=attrib)
    elem.extend(_build_sense_relation(rel) for rel in sense.relations)
    elem.extend(_build_example(ex) for ex in sense.examples)
    elem.extend(_build_count(cnt) for cnt in sense.counts)
    return elem


def _build_sense_relation(relation: SenseRelation) -> ET.Element:
    attrib = {'target': relation.target, 'relType': relation.type}
    attrib.update(_meta_dict(relation.meta))
    return ET.Element('SenseRelation', attrib=attrib)


def _build_example(example: Example) -> ET.Element:
    elem = ET.Element('Example')
    elem.text = example.text
    if example.language:
        elem.set('language', example.language)
    return elem


def _build_count(count: Count) -> ET.Element:
    elem = ET.Element('Count', attrib=_meta_dict(count.meta))
    elem.text = str(count.value)
    return elem


def _dump_synset(synset: Synset, out: TextIO, version: str) -> None:
    attrib = {'id': synset.id, 'ili': synset.ili}
    if synset.pos:
        attrib['partOfSpeech'] = synset.pos
    attrib.update(_meta_dict(synset.meta))
    if not synset.lexicalized:
        attrib['lexicalized'] = 'false'
    elem = ET.Element('Synset', attrib=attrib)
    elem.extend(_build_definition(defn) for defn in synset.definitions)
    if synset.ili_definition:
        elem.append(_build_ili_definition(synset.ili_definition))
    elem.extend(_build_synset_relation(rel) for rel in synset.relations)
    elem.extend(_build_example(ex) for ex in synset.examples)
    print(_tostring(elem, 2), file=out)


def _build_definition(definition: Definition) -> ET.Element:
    attrib = {}
    if definition.language:
        attrib['language'] = definition.language
    if definition.source_sense:
        attrib['sourceSense'] = definition.source_sense
    attrib.update(_meta_dict(definition.meta))
    elem = ET.Element('Definition', attrib=attrib)
    elem.text = definition.text
    return elem


def _build_ili_definition(ili_definition: ILIDefinition) -> ET.Element:
    elem = ET.Element('ILIDefinition', attrib=_meta_dict(ili_definition.meta))
    elem.text = ili_definition.text
    return elem


def _build_synset_relation(relation: SynsetRelation) -> ET.Element:
    attrib = {'target': relation.target, 'relType': relation.type}
    attrib.update(_meta_dict(relation.meta))
    return ET.Element('SynsetRelation', attrib=attrib)


def _dump_syntactic_behaviour(
        syntactic_behaviour: SyntacticBehaviour, out: TextIO, version: str
) -> None:
    elem = _build_syntactic_behaviour(syntactic_behaviour)
    print('    ' + _tostring(elem, 2), file=out)


def _build_syntactic_behaviour(syntactic_behaviour: SyntacticBehaviour) -> ET.Element:
    elem = ET.Element('SyntacticBehaviour')
    # if getattr(syntactic_behaviour, 'id', None):
    #     elem.set('id', syntactic_behaviour.id)
    elem.set('subcategorizationFrame', syntactic_behaviour.frame)
    if syntactic_behaviour.senses:
        elem.set('senses', ' '.join(syntactic_behaviour.senses))
    return elem


def _tostring(
        elem: ET.Element, level: int, short_empty_elements: bool = True
) -> str:
    _indent(elem, level)
    return ('  ' * level) + ET.tostring(
        elem,
        encoding='unicode',
        short_empty_elements=short_empty_elements
    )


def _indent(elem: ET.Element, level: int) -> None:
    self_indent = '\n' + '  ' * level
    child_indent = self_indent + '  '
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = child_indent
        for child in elem[:-1]:
            _indent(child, level + 1)
            child.tail = child_indent
        _indent(elem[-1], level + 1)
        elem[-1].tail = self_indent


def _meta_dict(m: Optional[Metadata]) -> Dict:
    if m:
        d = {f'dc:{key}': val for key, val in zip(m._fields, m) if val is not None}
    else:
        d = {}
    return d



import argparse
from pathlib import Path
import logging

import wn


def _download(args):
    if args.index:
        wn.config.load_index(args.index)
    for target in args.target:
        wn.download(target, add=args.add)


def _lexicons(args):
    for lex in wn.lexicons(lang=args.lang, lexicon=args.lexicon):
        print('\t'.join((lex.id, lex.version, f'[{lex.language}]', lex.label)))


def _projects(args):
    for info in wn.projects():
        key = 'i'
        key += 'c' if info['cache'] else '-'
        # key += 'a' if False else '-'  # TODO: check if project is added to db
        print(
            '\t'.join((
                key,
                info['id'],
                info['version'],
                f"[{info['language']}]",
                info['label'],
            ))
        )


def _path_type(arg):
    return Path(arg)


def _file_path_type(arg):
    path = Path(arg)
    if not path.is_file():
        raise argparse.ArgumentTypeError(f'cannot file file: {arg}')
    return path


parser = argparse.ArgumentParser(
    prog='python3 -m wn',
    description="Manage Wn's wordnet data from the command line.",
)
parser.add_argument(
    '-V', '--version', action='version', version=f'Wn {wn.__version__}'
)
parser.add_argument(
    '-v', '--verbose', action='count', dest='verbosity', default=0,
    help='increase verbosity (can repeat: -vv, -vvv)'
)
parser.add_argument(
    '-d', '--dir',
    type=_path_type,
    help="data directory for Wn's database and cache",
)
parser.set_defaults(func=lambda _: parser.print_help())
sub_parsers = parser.add_subparsers(title='subcommands')


parser_download = sub_parsers.add_parser(
    'download',
    description="Download wordnets and add them to Wn's database.",
    help='download wordnets',
)
parser_download.add_argument(
    'target', nargs='+', help='project specifiers or URLs'
)
parser_download.add_argument(
    '--index', type=_file_path_type, help='project index to use for downloading'
)
parser_download.add_argument(
    '--no-add', action='store_false', dest='add',
    help='download and cache without adding to the database'
)
parser_download.set_defaults(func=_download)


parser_lexicons = sub_parsers.add_parser(
    'lexicons',
    description="Display a list of installed lexicons.",
    help='list installed lexicons',
)
parser_lexicons.add_argument(
    '-l', '--lang', help='BCP 47 language code'
)
parser_lexicons.add_argument(
    '--lexicon', help='lexicon specifiers'
)
parser_lexicons.set_defaults(func=_lexicons)


parser_projects = sub_parsers.add_parser(
    'projects',
    description=(
        "Display a list of known projects. The first column shows the "
        "status for a project (i=indexed, c=cached)."
    ),
    help='list known projects',
)
parser_projects.set_defaults(func=_projects)


args = parser.parse_args()

logging.basicConfig(level=logging.ERROR - (min(args.verbosity, 3) * 10))

if args.dir:
    wn.config.data_directory = args.dir

args.func(args)



"""
Storage back-end interface.
"""

from typing import Dict
from pathlib import Path
import json
import sqlite3
import logging

import wn
from wn._types import Metadata, AnyPath
from wn._util import resources, short_hash
from wn import constants
from wn import lmf


logger = logging.getLogger('wn')


# Module Constants

DEBUG = False
NON_ROWID = 0  # imaginary rowid of non-existent row

# This stores hashes of the schema to check for version differences.
# When the schema changes, the hash will change. If the new hash is
# not added here, the 'test_schema_compatibility' test will fail. It
# is the developer's responsibility to only add compatible schema
# hashes here. If the schema change is not backwards-compatible, then
# clear all old hashes and only put the latest hash here. A hash can
# be generated like this:
#
# >>> import sqlite3
# >>> import wn
# >>> conn = sqlite3.connect(wn.config.database_path)
# >>> wn._db.schema_hash(conn)
#
COMPATIBLE_SCHEMA_HASHES = {
    '0cbec124b988d08e428b80d2b749563c2dccfa65',
}


# Optional metadata is stored as a JSON string

def _adapt_metadata(meta: lmf.Metadata) -> bytes:
    d = {key: val for key, val in zip(meta._fields, meta) if val is not None}
    return json.dumps(d).encode('utf-8')


def _convert_metadata(s: bytes) -> Metadata:  # note: wn._types.Metadata
    return json.loads(s)


def _convert_boolean(s: bytes) -> bool:
    return bool(int(s))


sqlite3.register_adapter(lmf.Metadata, _adapt_metadata)
sqlite3.register_converter('meta', _convert_metadata)
sqlite3.register_converter('boolean', _convert_boolean)


# The pool is a cache of open connections. Unless the database path is
# changed, there should only be zero or one.
pool: Dict[AnyPath, sqlite3.Connection] = {}


# The connect() function should be used for all connections

def connect() -> sqlite3.Connection:
    dbpath = wn.config.database_path
    if dbpath not in pool:
        initialized = dbpath.is_file()
        conn = sqlite3.connect(
            str(dbpath),
            detect_types=sqlite3.PARSE_DECLTYPES,
            check_same_thread=not wn.config.allow_multithreading,
        )
        # foreign key support needs to be enabled for each connection
        conn.execute('PRAGMA foreign_keys = ON')
        if DEBUG:
            conn.set_trace_callback(print)
        if not initialized:
            logger.info('initializing database: %s', dbpath)
            _initialize(conn)
        _check_schema_compatibility(conn, dbpath)

        pool[dbpath] = conn
    return pool[dbpath]


def _initialize(conn: sqlite3.Connection) -> None:
    schema = resources.read_text('wn', 'schema.sql')
    with conn:
        conn.executescript(schema)
        # prepare lookup tables
        conn.executemany(
            'INSERT INTO parts_of_speech (pos) VALUES (?)',
            ((pos,) for pos in constants.PARTS_OF_SPEECH))
        conn.executemany(
            'INSERT INTO adjpositions (position) VALUES (?)',
            ((adj,) for adj in constants.ADJPOSITIONS))
        conn.executemany(
            'INSERT INTO synset_relation_types (type) VALUES (?)',
            ((typ,) for typ in constants.SYNSET_RELATIONS))
        conn.executemany(
            'INSERT INTO sense_relation_types (type) VALUES (?)',
            ((typ,) for typ in constants.SENSE_RELATIONS))
        conn.executemany(
            'INSERT INTO lexicographer_files (id, name) VALUES (?,?)',
            ((id, name) for name, id in constants.LEXICOGRAPHER_FILES.items()))


def _check_schema_compatibility(conn: sqlite3.Connection, dbpath: Path) -> None:
    hash = schema_hash(conn)

    # if the hash is known, then we're all good here
    if hash in COMPATIBLE_SCHEMA_HASHES:
        return

    # otherwise, try to raise a helpful error message
    msg = ("Wn's schema has changed and is no longer compatible with the "
           f"database. Please move or delete {dbpath} and rebuild it.")
    try:
        specs = conn.execute('SELECT id, version FROM lexicons').fetchall()
    except sqlite3.OperationalError as exc:
        raise wn.Error(msg) from exc
    else:
        if specs:
            installed = '\n  '.join(f'{id}:{ver}' for id, ver in specs)
            msg += f" Lexicons currently installed:\n  {installed}"
        else:
            msg += ' No lexicons are currently installed.'
        raise wn.Error(msg)


def schema_hash(conn: sqlite3.Connection) -> str:
    query = 'SELECT sql FROM sqlite_master WHERE NOT sql ISNULL'
    schema = '\n\n'.join(row[0] for row in conn.execute(query))
    return short_hash(schema)



"""
Database retrieval queries.
"""

from typing import (
    Optional, Any, Dict, Set, List, Tuple, Collection, Iterator, Sequence
)
import itertools
import sqlite3

import wn
from wn._types import Metadata
from wn._db import connect, NON_ROWID


# Local Types

_Form = Tuple[
    str,            # form
    Optional[str],  # script
    int             # rowid
]
_Word = Tuple[
    str,          # id
    str,          # pos
    List[_Form],  # forms
    int,          # lexid
    int,          # rowid
]
_Synset = Tuple[
    str,  # id
    str,  # pos
    str,  # ili
    int,  # lexid
    int,  # rowid
]
_Synset_Relation = Tuple[str, int, str, str, str, int, int]  # relname, relid, *_Synset
_Sense = Tuple[
    str,  # id
    str,  # entry_id
    str,  # synset_id
    int,  # lexid
    int,  # rowid
]
_Sense_Relation = Tuple[str, int, str, str, str, int, int]  # relname, relid,  *_Sense
_Lexicon = Tuple[
    int,       # rowid
    str,       # id
    str,       # label
    str,       # language
    str,       # email
    str,       # license
    str,       # version
    str,       # url
    str,       # citation
    Metadata,  # metadata
]


def find_lexicons(
    lexicon: str,
    lang: str = None,
) -> Iterator[_Lexicon]:
    conn = connect()
    rows = conn.execute('SELECT rowid, id, version, language FROM lexicons').fetchall()
    rowids = _get_lexicon_rowids_for_lang(rows, lang)
    # the next call is somewhat expensive, so try to skip it in a common case
    if lexicon != '*':
        rowids &= _get_lexicon_rowids_for_lexicon(rows, lexicon)
    if rows and not rowids:
        raise wn.Error(
            f'no lexicon found with lang={lang!r} and lexicon={lexicon!r}'
        )
    for rowid in sorted(rowids):
        yield _get_lexicon(conn, rowid)


def get_lexicon(rowid: int) -> _Lexicon:
    conn = connect()
    return _get_lexicon(conn, rowid)


def _get_lexicon(conn: sqlite3.Connection, rowid: int) -> _Lexicon:
    query = '''
        SELECT DISTINCT rowid, id, label, language, email, license,
                        version, url, citation
        FROM lexicons
        WHERE rowid = ?
    '''
    row: Optional[_Lexicon] = conn.execute(query, (rowid,)).fetchone()
    if row is None:
        raise LookupError(rowid)  # should we have a WnLookupError?
    return row


def _get_lexicon_rowids_for_lang(
        rows: List[Tuple[int, str, str, str]], lang: Optional[str]
) -> Set[int]:
    lg_match: Set[int] = set()
    if lang:
        lg_match.update(rowid for rowid, _, _, language in rows if language == lang)
        if not lg_match:
            raise wn.Error(f"no lexicon found with language code '{lang}'")
    else:
        lg_match.update(row[0] for row in rows)
    return lg_match


def _get_lexicon_rowids_for_lexicon(
    rows: List[Tuple[int, str, str, str]],
    lexicon: str,
) -> Set[int]:
    lexmap: Dict[str, Dict[str, int]] = {}
    for rowid, id, version, _ in rows:
        lexmap.setdefault(id, {})[version] = rowid

    lex_match: Set[int] = set()
    for id_ver in lexicon.split():
        id, _, ver = id_ver.partition(':')

        if id == '*':
            for vermap in lexmap.values():
                for version, rowid in vermap.items():
                    if ver in ('', '*', version):
                        lex_match.add(rowid)

        elif id in lexmap:
            if ver == '*':
                lex_match.update(rowid for rowid in lexmap[id].values())
            elif ver == '':
                lex_match.add(max(lexmap[id].values()))  # last installed version
            elif ver in lexmap[id]:
                lex_match.add(lexmap[id][ver])
            else:
                raise wn.Error(f"no lexicon with id '{id}' found with version '{ver}'")

        else:
            raise wn.Error(f"no lexicon found with id '{id}'")

    return lex_match


def find_entries(
        id: str = None,
        form: str = None,
        pos: str = None,
        lexicon_rowids: Sequence[int] = None,
) -> Iterator[_Word]:
    conn = connect()
    query_parts = [
        'SELECT DISTINCT e.lexicon_rowid, e.rowid, e.id, p.pos,'
        '                f.form, f.script, f.rowid',
        '  FROM entries AS e',
        '  JOIN parts_of_speech AS p ON p.rowid = e.pos_rowid',
        '  JOIN forms AS f ON f.entry_rowid = e.rowid',
    ]

    params: Dict[str, Any] = {'id': id, 'form': form, 'pos': pos}
    conditions = []
    if id:
        conditions.append('e.id = :id')
    if form:
        conditions.append('e.rowid IN'
                          ' (SELECT entry_rowid FROM forms WHERE form = :form)')
    if pos:
        conditions.append('p.pos = :pos')
    if lexicon_rowids:
        kws = {f'lex{i}': rowid for i, rowid in enumerate(lexicon_rowids, 1)}
        params.update(kws)
        conditions.append(f'e.lexicon_rowid IN ({_kws(kws)})')

    if conditions:
        query_parts.append(' WHERE ' + '\n   AND '.join(conditions))

    query_parts.append(' ORDER BY e.rowid, e.id, f.rank')

    query = '\n'.join(query_parts)
    rows: Iterator[
        Tuple[int, int, str, str, str, Optional[str], int]
    ] = conn.execute(query, params)
    groupby = itertools.groupby
    for key, group in groupby(rows, lambda row: row[0:4]):
        lexid, rowid, id, pos = key
        forms = [(row[4], row[5], row[6]) for row in group]
        yield (id, pos, forms, lexid, rowid)


def find_senses(
        id: str = None,
        form: str = None,
        pos: str = None,
        lexicon_rowids: Sequence[int] = None,
) -> Iterator[_Sense]:
    conn = connect()
    query_parts = [
        'SELECT DISTINCT s.id, e.id, ss.id, s.lexicon_rowid, s.rowid'
        '  FROM senses AS s'
        '  JOIN entries AS e ON e.rowid = s.entry_rowid'
        '  JOIN synsets AS ss ON ss.rowid = s.synset_rowid'
    ]

    params: Dict[str, Any] = {'id': id, 'form': form, 'pos': pos}
    conditions = []
    if id:
        conditions.append('s.id = :id')
    if form:
        conditions.append('s.entry_rowid IN'
                          ' (SELECT entry_rowid FROM forms WHERE form = :form)')
    if pos:
        conditions.append('e.pos_rowid IN'
                          ' (SELECT p.rowid'
                          '    FROM parts_of_speech AS p'
                          '   WHERE p.pos = :pos)')
    if lexicon_rowids:
        kws = {f'lex{i}': rowid for i, rowid in enumerate(lexicon_rowids, 1)}
        params.update(kws)
        conditions.append(f's.lexicon_rowid IN ({_kws(kws)})')

    if conditions:
        query_parts.append(' WHERE ' + '\n   AND '.join(conditions))

    query = '\n'.join(query_parts)
    rows: Iterator[_Sense] = conn.execute(query, params)
    yield from rows


def find_synsets(
        id: str = None,
        form: str = None,
        pos: str = None,
        ili: str = None,
        lexicon_rowids: Sequence[int] = None,
) -> Iterator[_Synset]:
    conn = connect()
    query_parts = [
        'SELECT DISTINCT ss.id, p.pos, ss.ili, ss.lexicon_rowid, ss.rowid',
        '  FROM synsets AS ss',
        '  JOIN parts_of_speech AS p ON p.rowid = ss.pos_rowid',
    ]

    params: Dict[str, Any] = {'id': id, 'form': form, 'pos': pos, 'ili': ili}
    conditions = []
    if id:
        conditions.append('ss.id = :id')
    if form:
        query_parts.append(
            '  JOIN (SELECT _s.synset_rowid, _s.entry_rowid, _s.entry_rank'
            '          FROM senses AS _s'
            '          JOIN forms AS f'
            '            ON f.entry_rowid = _s.entry_rowid'
            '         WHERE f.form = :form) AS s'
            '    ON s.synset_rowid = ss.rowid'
        )
    if pos:
        conditions.append('p.pos = :pos')
    if ili:
        conditions.append('ss.ili = :ili')
    if lexicon_rowids:
        kws = {f'lex{i}': rowid for i, rowid in enumerate(lexicon_rowids, 1)}
        params.update(kws)
        conditions.append(f'ss.lexicon_rowid IN ({_kws(kws)})')

    if conditions:
        query_parts.append(' WHERE ' + '\n   AND '.join(conditions))

    if form:
        query_parts.append(' ORDER BY s.entry_rowid, s.entry_rank')

    query = '\n'.join(query_parts)
    rows: Iterator[_Synset] = conn.execute(query, params)
    yield from rows


def get_synsets_for_ilis(
        ilis: Collection[str],
        lexicon_rowids: Sequence[int],
) -> Iterator[_Synset]:
    conn = connect()
    query = f'''
        SELECT DISTINCT ss.id, p.pos, ss.ili, ss.lexicon_rowid, ss.rowid
          FROM synsets as ss
          JOIN parts_of_speech AS p
            ON p.rowid = ss.pos_rowid
         WHERE ss.ili IN ({_qs(ilis)})
           AND ss.lexicon_rowid IN ({_qs(lexicon_rowids)})
    '''
    params = *ilis, *lexicon_rowids
    result_rows: Iterator[_Synset] = conn.execute(query, params)
    yield from result_rows


def get_synset_relations(
        source_rowids: Collection[int],
        relation_types: Collection[str],
) -> Iterator[_Synset_Relation]:
    conn = connect()
    params: List = []
    constraint = ''
    if '*' not in relation_types:
        constraint = f'WHERE t.type IN ({_qs(relation_types)})'
        params.extend(relation_types)
    params.extend(source_rowids)
    query = f'''
          WITH relation_types(type_rowid, type) AS  -- relation type lookup
               (SELECT t.rowid, t.type
                  FROM synset_relation_types AS t
                {constraint})
        SELECT DISTINCT rel.type, rel.rowid,
                        tgt.id, p.pos, tgt.ili,
                        tgt.lexicon_rowid, tgt.rowid
          FROM (SELECT type, target_rowid, srel.rowid
                  FROM synset_relations AS srel
                  JOIN relation_types USING (type_rowid)
                 WHERE source_rowid IN ({_qs(source_rowids)})) AS rel
          JOIN synsets AS tgt
            ON tgt.rowid = rel.target_rowid
          JOIN parts_of_speech AS p
            ON p.rowid = tgt.pos_rowid
    '''
    result_rows: Iterator[_Synset_Relation] = conn.execute(query, params)
    yield from result_rows


def get_definitions(synset_rowid: int) -> List[Tuple[str, str, int]]:
    conn = connect()
    query = '''
        SELECT definition, language, rowid
          FROM definitions
         WHERE synset_rowid = ?
    '''
    return conn.execute(query, (synset_rowid,)).fetchall()


_SANITIZED_EXAMPLE_PREFIXES = {
    'senses': 'sense',
    'synsets': 'synset',
}


def get_examples(rowid: int, table: str) -> List[Tuple[str, str, int]]:
    conn = connect()
    prefix = _SANITIZED_EXAMPLE_PREFIXES.get(table)
    if prefix is None:
        raise wn.Error(f"'{table}' does not have examples")
    query = f'''
        SELECT example, language, rowid
          FROM {prefix}_examples
         WHERE {prefix}_rowid = ?
    '''
    return conn.execute(query, (rowid,)).fetchall()


def _get_senses(rowid: int, sourcetype: str) -> Iterator[_Sense]:
    conn = connect()
    query = f'''
        SELECT s.id, e.id, ss.id, s.lexicon_rowid, s.rowid
          FROM senses AS s
          JOIN entries AS e
            ON e.rowid = s.entry_rowid
          JOIN synsets AS ss
            ON ss.rowid = s.synset_rowid
         WHERE s.{sourcetype}_rowid = ?
    '''
    return conn.execute(query, (rowid,))


def get_entry_senses(rowid: int) -> Iterator[_Sense]:
    yield from _get_senses(rowid, 'entry')


def get_synset_members(rowid: int) -> Iterator[_Sense]:
    yield from _get_senses(rowid, 'synset')


def get_sense_relations(
        source_rowid: int,
        relation_types: Collection[str],
) -> Iterator[_Sense_Relation]:
    conn = connect()
    params: List = []
    constraint = ''
    if '*' not in relation_types:
        constraint = f'WHERE t.type IN ({_qs(relation_types)})'
        params.extend(relation_types)
    params.append(source_rowid)
    query = f'''
          WITH relation_types(type_rowid, type) AS  -- relation type lookup
               (SELECT t.rowid, t.type
                  FROM sense_relation_types AS t
                {constraint})
        SELECT DISTINCT rel.type, rel.rowid,
                        s.id, e.id, ss.id,
                        s.lexicon_rowid, s.rowid
          FROM (SELECT type, target_rowid, srel.rowid
                  FROM sense_relations AS srel
                  JOIN relation_types USING (type_rowid)
                 WHERE source_rowid = ?) AS rel
          JOIN senses AS s
            ON s.rowid = rel.target_rowid
          JOIN entries AS e
            ON e.rowid = s.entry_rowid
          JOIN synsets AS ss
            ON ss.rowid = s.synset_rowid
    '''
    rows: Iterator[_Sense_Relation] = conn.execute(query, params)
    yield from rows


def get_sense_synset_relations(
        source_rowid: int,
        relation_types: Collection[str],
) -> Iterator[_Synset_Relation]:
    conn = connect()
    params: List = [source_rowid]
    constraint = ''
    if '*' not in relation_types:
        constraint = f'WHERE t.type IN ({_qs(relation_types)})'
        params.extend(relation_types)
    query = f'''
          WITH relation_types(type_rowid, type) AS  -- relation type lookup
               (SELECT t.rowid, t.type
                  FROM sense_relation_types AS t
                {constraint})
        SELECT DISTINCT rel.type, rel.rowid,
                        ss.id, p.pos, ss.ili,
                        ss.lexicon_rowid, ss.rowid
          FROM (SELECT type, target_rowid, srel.rowid
                  FROM sense_synset_relations AS srel
                  JOIN relation_types USING (type_rowid)
                 WHERE source_rowid = ?) AS rel
          JOIN synsets AS ss
            ON ss.rowid = rel.target_rowid
          JOIN parts_of_speech AS p
            ON p.rowid = ss.pos_rowid
    '''
    rows: Iterator[_Synset_Relation] = conn.execute(query, params)
    yield from rows


_SANITIZED_METADATA_TABLES = {
    'lexicons': 'lexicons',
    'entries': 'entries',
    'senses': 'senses',
    'synsets': 'synsets',
    'sense_relations': 'sense_relations',
    'sense_synset_relations': 'sense_synset_relations',
    'synset_relations': 'synset_relations',
    'sense_examples': 'sense_examples',
    'synset_examples': 'synset_examples',
    'definitions': 'definitions',
}


def get_metadata(
    rowid: int, table: str
) -> Metadata:
    conn = connect()
    tablename = _SANITIZED_METADATA_TABLES.get(table)
    if tablename is None:
        raise wn.Error(f"'{table}' does not contain metadata")
    query = f'SELECT metadata FROM {tablename} WHERE rowid=?'
    return conn.execute(query, (rowid,)).fetchone()[0] or {}


_SANITIZED_LEXICALIZED_TABLES = {
    'senses': 'senses',
    'synsets': 'synsets',
}


def get_lexicalized(rowid: int, table: str) -> bool:
    conn = connect()
    tablename = _SANITIZED_LEXICALIZED_TABLES.get(table)
    if tablename is None:
        raise wn.Error(f"'{table}' does not mark lexicalization")
    if rowid == NON_ROWID:
        return False
    query = f'SELECT lexicalized FROM {tablename} WHERE rowid=?'
    return conn.execute(query, (rowid,)).fetchone()[0]


def _qs(xs: Collection) -> str: return ','.join('?' * len(xs))
def _kws(xs: Collection) -> str: return ','.join(f':{x}' for x in xs)




"""
Wordnet Interface.
"""

__all__ = (
    '__version__',
    'Wordnet',
    'download',
    'add',
    'remove',
    'export',
    'projects',
    'lexicons',
    'Lexicon',
    'word',
    'words',
    'Word',
    'Form',
    'sense',
    'senses',
    'Sense',
    'synset',
    'synsets',
    'Synset',
    'Error',
)

from wn._meta import __version__
from wn._exceptions import Error
from wn._config import config  # noqa: F401
from wn._add import add, remove
from wn._export import export
from wn._download import download
from wn._core import (
    projects,
    lexicons, Lexicon,
    word, words, Word, Form,
    sense, senses, Sense,
    synset, synsets, Synset,
    Wordnet
)




from typing import Union, Mapping, Sequence, Dict, Any
from pathlib import Path

# For functions taking a filesystem path as a str or a pathlib.Path
AnyPath = Union[str, Path]

# Synset and Sense relations map a relation type to one or more ids
RelationMap = Mapping[str, Sequence[str]]

# User-facing metadata representation
Metadata = Dict[str, Any]



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['wn']

package_data = \
{'': ['*']}

install_requires = \
['requests~=2.25', 'toml~=0.10']

extras_require = \
{":python_version < '3.7'": ['importlib_resources'],
 'doc': ['sphinx', 'furo', 'sphinx-copybutton'],
 'test': ['pytest', 'mypy', 'flake8', 'nox']}

setup(name='wn',
      version='0.5.0',
      description='Wordnet Interface.',
      author='Michael Wayne Goodman',
      author_email='goodman.m.w@gmail.com',
      url='https://github.com/goodmami/wn',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )



import nox


@nox.session
def lint(session):
    session.install('flake8', 'mypy')
    session.install('.')
    session.run('flake8', '--max-line-length', '88', 'wn')
    session.run('mypy', 'wn')


@nox.session
def test(session):
    session.install('pytest')
    session.install('.')
    session.run('pytest')




import pytest

import wn


@pytest.mark.usefixtures('mini_db')
def test_word_derived_words():
    assert len(wn.word('test-en-example-n').derived_words()) == 1
    assert len(wn.word('test-es-ejemplo-n').derived_words()) == 1


@pytest.mark.usefixtures('mini_db')
def test_synset_hypernyms():
    assert wn.synset('test-en-0002-n').hypernyms() == [
        wn.synset('test-en-0001-n')
    ]
    assert wn.synset('test-en-0001-n').hypernyms() == []


@pytest.mark.usefixtures('mini_db')
def test_synset_hypernyms_expand_default():
    assert wn.synset('test-es-0002-n').hypernyms() == [
        wn.synset('test-es-0001-n')
    ]
    assert wn.synset('test-es-0001-n').hypernyms() == []


@pytest.mark.usefixtures('mini_db')
def test_synset_hypernyms_expand_empty():
    w = wn.Wordnet(lang='es', expand='')
    assert w.synset('test-es-0002-n').hypernyms() == []


@pytest.mark.usefixtures('mini_db')
def test_synset_hypernyms_expand_specified():
    w = wn.Wordnet(lang='es', expand='test-en')
    assert w.synset('test-es-0002-n').hypernyms() == [
        wn.synset('test-es-0001-n')
    ]



"""Wn utility classes."""

from typing import TextIO
import sys


class ProgressHandler:
    """An interface for updating progress in long-running processes.

    Long-running processes in Wn, such as :func:`wn.download` and
    :func:`wn.add`, call to a progress handler object as they go.  The
    default progress handler used by Wn is :class:`ProgressBar`, which
    updates progress by formatting and printing a textual bar to
    stderr. The :class:`ProgressHandler` class may be used directly,
    which does nothing, or users may create their own subclasses for,
    e.g., updating a GUI or some other handler.

    The initialization parameters, except for ``file``, are stored in
    a :attr:`kwargs` member and may be updated after the handler is
    created through the :meth:`set` method. The :meth:`update` method
    is the primary way a counter is updated. The :meth:`flash` method
    is sometimes called for simple messages. When the process is
    complete, the :meth:`close` method is called, optionally with a
    message.

    """

    def __init__(
        self,
        *,
        message: str = '',
        count: int = 0,
        total: int = 0,
        unit: str = '',
        status: str = '',
        file: TextIO = sys.stderr,
    ):
        self.file = file
        self.kwargs = {
            'count': count,
            'total': total,
            'message': message,
            'unit': unit,
            'status': status,
        }

    def update(self, n: int = 1) -> None:
        """Update the counter with the increment value *n*.

        This method should update the ``count`` key of :attr:`kwargs`
        with the increment value *n*. After this, it is expected to
        update some user-facing progress indicator.

        """
        self.kwargs['count'] += n  # type: ignore

    def set(self, **kwargs) -> None:
        """Update progress handler parameters.

        Calling this method also runs :meth:`update` with an increment
        of 0, which causes a refresh of any indicator without changing
        the counter.

        """
        self.kwargs.update(**kwargs)
        self.update(0)

    def flash(self, message: str) -> None:
        """Issue a message unrelated to the current counter.

        This may be useful for multi-stage processes to indicate the
        move to a new stage, or to log unexpected situations.

        """
        pass

    def close(self) -> None:
        """Close the progress handler.

        This might be useful for closing file handles or cleaning up
        resources.

        """
        pass


class ProgressBar(ProgressHandler):
    """A :class:`ProgressHandler` subclass for printing a progress bar.

    Example:
        >>> p = ProgressBar(message='Progress: ', total=10, unit='units')
        >>> p.update(3)
        Progress: [#########                     ] (3/10 units)

    """

    FMT = '\r{message} [{fill:<{width}}] ({count}/{total} {unit}) {status}'

    def update(self, n: int = 1) -> None:
        self.kwargs['count'] += n  # type: ignore
        s = self.format()
        if self.file:
            print('\r\033[K', end='', file=self.file)
            print(s, end='', file=self.file)

    def set(self, **kwargs) -> None:
        self.kwargs.update(**kwargs)
        self.update(0)

    def format(self) -> str:
        _kw = self.kwargs
        width = 30
        total: int = _kw['total']  # type: ignore
        count: int = _kw['count']  # type: ignore

        if total > 0:
            num = min(count, total) * width
            fill = (num // total) * '#'
            part = ((num % total) * 3) // total
            if part:
                fill += '-='[part-1]
        else:
            fill = '-' * width

        return self.FMT.format(fill=fill, width=width, **_kw)

    def flash(self, message: str) -> None:
        print(f'\r\033[K{message}', end='', file=self.file)

    def close(self) -> None:
        print(file=self.file)




from typing import List, Set, Sequence

import wn
from wn._types import AnyPath
from wn import lmf
from wn._queries import (
    find_entries,
    find_senses,
    find_synsets,
    get_entry_senses,
    get_sense_relations,
    get_sense_synset_relations,
    get_synset_relations,
    get_examples,
    get_definitions,
    get_metadata,
    get_lexicalized,
)
from wn._core import Lexicon


def export(
        lexicons: Sequence[Lexicon], destination: AnyPath, version: str = '1.0'
) -> None:
    """Export lexicons from the database to a WN-LMF file.

    More than one lexicon may be exported in the same file, subject to
    these conditions:

    - identifiers on wordnet entities must be unique in all lexicons
    - lexicons extensions may not be exported with their dependents

    >>> w = wn.Wordnet(lexicon='cmnwn zsmwn')
    >>> wn.export(w.lexicons(), 'cmn-zsm.xml')

    Args:
        lexicons: sequence of :class:`wn.Lexicon` objects
        destination: path to the destination file
        version: LMF version string

    """
    _precheck(lexicons)
    _lexicons = [_export_lexicon(lex) for lex in lexicons]
    lmf.dump(_lexicons, destination, version=version)


def _precheck(lexicons: Sequence[Lexicon]) -> None:
    all_ids: Set[str] = set()
    for lex in lexicons:
        lexids = (lex._id,)
        idset = {lex.id}
        idset.update(row[0] for row in find_entries(lexicon_rowids=lexids))
        idset.update(row[0] for row in find_senses(lexicon_rowids=lexids))
        idset.update(row[0] for row in find_synsets(lexicon_rowids=lexids))
        # TODO: syntactic behaviours
        if all_ids.intersection(idset):
            raise wn.Error('cannot export: non-unique identifiers in lexicons')
        all_ids |= idset


def _export_lexicon(lexicon: Lexicon) -> lmf.Lexicon:
    lexids = (lexicon._id,)
    return lmf.Lexicon(
        id=lexicon.id,
        label=lexicon.label,
        language=lexicon.language,
        email=lexicon.email,
        license=lexicon.license,
        version=lexicon.version,
        url=lexicon.url or '',
        citation=lexicon.citation or '',
        lexical_entries=_export_lexical_entries(lexids),
        synsets=_export_synsets(lexids),
        syntactic_behaviours=_export_syntactic_behaviours(lexids),
        meta=lmf.Metadata(**lexicon.metadata())
    )


def _export_lexical_entries(lexids: Sequence[int]) -> List[lmf.LexicalEntry]:
    LexicalEntry = lmf.LexicalEntry
    Lemma = lmf.Lemma
    Form = lmf.Form
    return [
        LexicalEntry(
            id,
            Lemma(forms[0][0], pos, script=forms[0][1] or ''),  # TODO: tags
            forms=[Form(form, script or '')
                   for form, script, _ in forms[1:]],  # TODO: tags
            senses=_export_senses(rowid),
            meta=_export_metadata(rowid, 'entries'),
        )
        for id, pos, forms, _, rowid
        in find_entries(lexicon_rowids=lexids)
    ]


def _export_senses(entry_rowid: int) -> List[lmf.Sense]:
    Sense = lmf.Sense
    return [
        Sense(
            id,
            synset,
            relations=_export_sense_relations(rowid),
            examples=_export_examples(rowid, 'senses'),
            # TODO: counts
            lexicalized=get_lexicalized(rowid, 'senses'),
            # TODO: adjposition
            meta=_export_metadata(rowid, 'senses'),
        )
        for id, _, synset, _, rowid
        in get_entry_senses(entry_rowid)
    ]


def _export_sense_relations(sense_rowid: int) -> List[lmf.SenseRelation]:
    SenseRelation = lmf.SenseRelation
    relations = [
        SenseRelation(
            id,
            type,
            meta=_export_metadata(rowid, 'sense_relations')
        )
        for type, rowid, id, *_
        in get_sense_relations(sense_rowid, '*')
    ]
    relations.extend(
        SenseRelation(
            id,
            type,
            meta=_export_metadata(rowid, 'sense_synset_relations')
        )
        for type, rowid, id, *_
        in get_sense_synset_relations(sense_rowid, '*')
    )
    return relations


def _export_examples(rowid: int, table: str) -> List[lmf.Example]:
    Example = lmf.Example
    return [
        Example(
            text,
            language,
            meta=_export_metadata(rowid, f'{table[:-1]}_examples')
        )
        for text, language, rowid
        in get_examples(rowid, table)
    ]


def _export_synsets(lexids: Sequence[int]) -> List[lmf.Synset]:
    Synset = lmf.Synset
    return [
        Synset(
            id,
            ili or '',
            pos,
            definitions=_export_definitions(rowid),
            # TODO: ili_definition,
            relations=_export_synset_relations(rowid),
            examples=_export_examples(rowid, 'synsets'),
            lexicalized=get_lexicalized(rowid, 'synsets'),
            meta=_export_metadata(rowid, 'synsets'),
        )
        for id, pos, ili, _, rowid
        in find_synsets(lexicon_rowids=lexids)
    ]


def _export_definitions(rowid: int) -> List[lmf.Definition]:
    Definition = lmf.Definition
    return [
        Definition(
            text,
            language,
            meta=_export_metadata(rowid, 'definitions')
        )
        for text, language, rowid
        in get_definitions(rowid)
    ]


def _export_synset_relations(synset_rowid: int) -> List[lmf.SynsetRelation]:
    SynsetRelation = lmf.SynsetRelation
    relations = [
        SynsetRelation(
            id,
            type,
            meta=_export_metadata(rowid, 'synset_relations')
        )
        for type, rowid, id, *_
        in get_synset_relations((synset_rowid,), '*')
    ]
    return relations


def _export_syntactic_behaviours(lexids: Sequence[int]) -> List[lmf.SyntacticBehaviour]:
    return []  # TODO


def _export_metadata(rowid: int, table: str) -> lmf.Metadata:
    return lmf.Metadata(**get_metadata(rowid, table))
